/****************************************************************************/
/****************************************************************************/

/*
   API Debugging class for iPhone.
   Copyright (c) 2008 Core Security Technologies, Core SDI Inc.
   Developed by Nicolas A. Economou (neconomou@corest.com).

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your option)
any later version. 

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
more details. 

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/****************************************************************************/
/****************************************************************************/

/* iphone_debug.cpp */

/****************************************************************************/
/****************************************************************************/

#include <ctype.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <mach/exception_types.h>
#include <mach/mach_init.h>
#include <mach/mach_port.h>
#include <mach/mach_traps.h>
#include <mach/task.h>
#include <mach/task_info.h>
#include <mach/thread_act.h>
#include <mach/thread_info.h>
#include <mach/vm_map.h>
#include <mach-o/loader.h>
#include <mach-o/nlist.h>
#include <sys/ptrace.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <mach/boolean.h>
#include <mach/message.h>

/* Include necesario para desensamblar */
#define IPHONE
#include "disasm.h"

/* Para guardar los simbolos */
#include <map>
#include <list>
#include <string>
using namespace std;

/****************************************************************************/
/****************************************************************************/

extern "C" boolean_t exc_server ( mach_msg_header_t * , mach_msg_header_t * );
extern "C" kern_return_t catch_exception_raise ( mach_port_t , mach_port_t , mach_port_t , exception_type_t , exception_data_t , mach_msg_type_number_t );
int get_file_entry_point ( char *filename , unsigned int *entry_point_in_file , unsigned int *entry_point_in_memory );

/****************************************************************************/
/****************************************************************************/

#define FALSE                                 0
#define TRUE                                  1

#define MODE_THUMB_ENABLED                    0x20

#define REGISTER_PC                           15
#define REGISTER_CTRL                         16

#define CANTIDAD_REGISTROS                    17
#define MAX_LEN_INSTRUCTION                   4

#define DEBUG_EXCEPTION_CREATE_PROCESS        0
#define DEBUG_EXCEPTION_ACCESS_VIOLATION      1
#define DEBUG_EXCEPTION_INVALID_INSTRUCTION   2
#define DEBUG_EXCEPTION_BREAKPOINT            6
#define DEBUG_EXCEPTION_EXIT_PROCESS          100

#define DEBUG_PROTECTION_READ                 1
#define DEBUG_PROTECTION_WRITE                2
#define DEBUG_PROTECTION_EXECUTE              4

/****************************************************************************/
/****************************************************************************/

typedef struct
{
  int enabled;
  int oneshot;
  int breakpoint_mode;
  int recuperador;
  int system;
  unsigned int address;
  char instruccion_original [ 4 ];
  unsigned int len_instruccion_original;
  unsigned int direccion_breakpoint_a_recuperar;
} Breakpoint;


// Esta estructura es para informar al usuario
// parte de la info de cada breakpoint

typedef struct
{
  unsigned int address;
  int enabled;
  int oneshot;
} BreakpointInfo;

/****************************************************************************/
/****************************************************************************/

char *register_names [] = {                                                               
                            "r0" , "r1" , "r2" , "r3",    \
                            "r4" , "r5" , "r6" , "r7",    \
                            "r8" , "r9" , "r10" , "r11",  \
                            "r12" , "sp" , "lr" , "pc",   \
                            "ctrl",
                          };

char *breakpoint_instruction [] = {
                                    "\x00\xbe",         // thumb mode
                                    "\x70\x00\x20\xe1", // normal mode
                                  };

char while_true_instruction [] = { 0xfe , 0xff , 0xff , 0xea };

unsigned int exception_thread;
unsigned int exception_type;
unsigned int exception_code;

/* Variables usadas para controlar el CTRL+C */
int running = FALSE;
int breakear = FALSE;

/****************************************************************************/
/****************************************************************************/

class Debug
{
  unsigned int direccion_breakpoint_a_recuperar;
  unsigned int inferior_thread_count;
  thread_array_t inferior_threads;
  int len_breakpoint_a_recuperar;
  mach_port_t exception_port;
  mach_port_name_t port;
  int pid;
  int tracear;

  // Variables para mantener el estado del listado de los simbolos
  map <string,unsigned int> :: iterator symbol;
  char libname [ 256 + 1 ];
  char funcname [ 256 + 1 ];
  int especific_lib;
  int especific_function;

  // Diccionario que contiene el nombre de las librerias
  map <string,bool> librerias;
  map <unsigned int,string> address_to_name;
  map <string,unsigned int> name_to_address;

  // Diccionario con la lista de brteakpoints
  map <unsigned int,Breakpoint *> breakpoint_list;

  // Lista donde voy a guardar parte de la info para que pueda ver el usuario cuando haga la consulta 
  list <BreakpointInfo *> user_breakpoint_list;

  private:
    int load_symbol_table ( char *path );

  public:
    Debug ();
    int Execute ( char *process_name , char **allargs );
    int Attach ( int pid );
    int LoadSymbols ( char *filename );
    int Continue ( void );
    int Wait ( void );
    int Close ( void );
    int GetPc ( void );
    int GetCurrentPid ( void );
    int GetCurrentTid ( void );
    int *GetThreadList ( int *cantidad );
    int SetBreakpoint ( unsigned int address );
    int UnsetBreakpoint ( unsigned int address );
    int SetOneShotBreakpoint ( unsigned int address );
    int EnableBreakpoint ( unsigned int address );
    int DisableBreakpoint ( unsigned int address );
    list <BreakpointInfo *> *GetBreakpointList ( void );
    int EnableTrace ( void );
    int DisableTrace ( void );
    int ReadMemory ( void *dest , void *source , unsigned int len , unsigned int *leidos );
    int WriteMemory ( void *dest , void *source , unsigned int len );
    int Search ( void *addr_base , void *addr_limite , char *pattern , unsigned int len , unsigned int *addr_found );
    int GetRegisters ( int thread , unsigned int *registers );
    int SetRegisters ( int thread , unsigned int *registers );
    int GetEventCode ( void );
    int GetExceptionAddress ( void );
    int ResolveAddressToName ( unsigned int address , char *name , int *offset );
    int ResolveNameToAddress ( char *name , unsigned int *address );
    int GetSymbolFrom ( char *lib_function , unsigned int *address , const char **psymbol );
    int GetMemoryAttributes ( unsigned int address , unsigned int *base , unsigned int *len , unsigned int *attributes );
    int SetMemoryAttributes ( unsigned int address , unsigned int len , unsigned int *old_attributes );
    int Disassembly ( unsigned int address , unsigned int *len , char *dest_buffer );
};

/****************************************************************************/
/****************************************************************************/

/* Esta funcion es el handler de excepciones que llama el kernel */

kern_return_t catch_exception_raise(mach_port_t exception_port, mach_port_t
    thread, mach_port_t task, exception_type_t exception, exception_data_t
    code, mach_msg_type_number_t code_count)
{
/* Seteo las variables de estado */
  exception_thread = thread;
  exception_type = exception;
  exception_code = ( unsigned int ) code;

/* Retorno OK */
  return KERN_SUCCESS;
}

/****************************************************************************/

// Handler para CTRL+C
void ctrl_c ( int value )
{
/* Si el proceso esta corriendo */
  if ( running == TRUE )
  {
  /* Mensaje al usuario */
    puts ( "break" );

  /* Marco la variable global para que el metodo Wait la use */
    breakear = TRUE;
  }
}

/****************************************************************************/

int get_file_entry_point ( char *filename , unsigned int *entry_point_in_file , unsigned int *entry_point_in_memory )
{
  FILE *f = NULL;
  struct mach_header mh;
  struct load_command lc;
  int pos;
  int i;
  char *buffer;
  int ret = FALSE;

/* Abro el archivo */
  f = fopen ( filename , "rb" );

/* Si no lo pude abrir */
  if ( f == NULL )
  {
    return ( ret );
  }

/* Levanto el header del file */
  fread ( &mh , sizeof ( struct mach_header ) , 1 , f );

/* Si no es un header valido */
  if ( mh.magic != MH_MAGIC )
  {
  /* Cierro el archivo */
    fclose ( f );

    return ( ret );
  }

/* Busco la seccion _TEXT */
  for ( i = 0 ; i < mh.ncmds ; i++ )
  {
  /* Levanto la siguiente seccion */
    fread ( &lc , sizeof ( struct load_command ), 1 , f );

  /* Si no es una seccion que tenga codigo o datos */
    if ( lc.cmd != 0x01 )
    {
    /* Avanzo a la proxima seccion */
      fseek ( f , lc.cmdsize - sizeof ( struct load_command ) , SEEK_CUR );
      continue;
    }

  /* Alloco memoria para levantar la seccion completa a memoria */
    buffer = ( char * ) malloc ( lc.cmdsize - sizeof ( struct load_command ) );

  /* Levanto la seccion completa a memoria */
    fread ( buffer , lc.cmdsize - sizeof ( struct load_command ) , 1 , f );

  /* Si es la seccion _TEXT */
    if ( strcmp ( buffer , "__TEXT" ) == 0 )
    {
    /* Busco la subseccion __text */
      for ( pos = 0 ; pos < ( lc.cmdsize - sizeof ( struct load_command ) - 6 ) ; pos ++ )
      {
      /* Si encontre la subseccion */
        if ( strstr ( &buffer [ pos ] , "__text" ) != NULL )
        {
        /* Direccion en memoria del ENTRY_POINT */
          *entry_point_in_memory = * ( unsigned int * ) &buffer [ pos + 0x20 ];
          *entry_point_in_file = * ( unsigned int * ) &buffer [ pos + 0x28 ];

        /* Libero la memoria allocada */
          free ( buffer );

        /* Cierro el archivo */
          fclose ( f );

        /* Retorno OK */
          return ( TRUE );
        }
      }
    }

  /* Libero la memoria allocada */
    free ( buffer );
  }

/* Cierro el archivo */
  fclose ( f );

/* Salgo */
  return ( ret );
}

/****************************************************************************/

int Debug::load_symbol_table ( char *path )
{
  FILE *f = NULL;
  char my_string [ 256 ];
  string function_name;
  struct load_command lc;
  struct mach_header mh;
  struct nlist sym;
  struct symtab_command sc;
  string libreria;
  char *function_names;
  int ret = TRUE;
  int i;

  printf ( "loading symbols from %s\n" , path );

/* Abro el archivo */
  f = fopen ( path , "rb" );

/* Si no lo pude abrir */
  if ( f == NULL )
  {
    return ( FALSE );
  }

/* Levanto el header del file */
  fread ( &mh , sizeof ( struct mach_header ) , 1 , f );

/* Si no es un header valido */
  if ( mh.magic != MH_MAGIC )
  {
    return ( FALSE );
  }

/* Busco la seccion de imports */
  for ( i = 0 ; i < mh.ncmds ; i++ )
  {
  /* Levanto la siguiente seccion */
    fread ( &lc , sizeof ( struct load_command ), 1 , f );

  /* Si no es la seccion que tiene los simbolos */
    if ( lc.cmd != 0x0c )
    {
    /* Avanzo a la proxima seccion */
      fseek ( f , lc.cmdsize - sizeof ( struct load_command ) , SEEK_CUR );
      continue;
    }

  /* Salteo 0x10 bytes del header de la seccion */
    fseek ( f , 0x10 , SEEK_CUR );

  /* Levanto el nombre de la lib importada */
    fread ( ( void * ) my_string , lc.cmdsize - sizeof ( struct load_command ) - 0x10 , 1 , f );

  /* nombre de la libreria a la que apunta el file en memoria */
    libreria = my_string;

  /* Si la libreria no fue cargada */
    if ( this -> librerias [ libreria ] == FALSE )
    {
    /* Cargo una nueva libreria */
      this -> librerias [ libreria ] = TRUE;

    /* Cargo los simbolos de la libreria */
      this -> load_symbol_table ( my_string );
    }
  }

/* Vuelvo a la posicion anterior */
  fseek ( f , sizeof ( struct mach_header ) , SEEK_SET );

/* Busco la tabla de symbolos */
  for ( i = 0 ; i < mh.ncmds ; i++ )
  {
  /* Levanto la siguiente seccion */
    fread ( &lc , sizeof ( struct load_command ), 1 , f );

  /* Si no es la seccion que tiene los simbolos */
    if ( lc.cmd != LC_SYMTAB )
    {
      fseek(f, lc.cmdsize - sizeof(struct load_command), SEEK_CUR);
      continue;
    }

    break;
  }

/* Si no pude encontrar la seccion de los simbolos */
  if ( lc.cmd != LC_SYMTAB )
  {
    return ( FALSE );
  }

/* Leo el header de la seccion de los simbolos */
  fseek ( f , -sizeof ( struct load_command ) , SEEK_CUR );
  fread ( &sc , sizeof ( struct symtab_command ) , 1 , f );

/* Me posiciono en la zona de nombres de funciones */
  function_names = ( char * ) malloc ( sc.strsize );
  fseek ( f , sc.stroff , SEEK_SET );

/* Levanto toda la zona de nombres a la memoria */
  fread ( function_names , 1 , sc.strsize , f );

/* Me posiciono en el area de los simbolos */
  fseek ( f , sc.symoff , SEEK_SET );

/* Recorro todas las entradas */
  for ( i = 0 ; i < sc.nsyms ; i ++ )
  {
  /* Levanto la siguiente entrada */
    fread ( &sym , sizeof ( struct nlist ) , 1 , f );

  /* Si no es una entrada valida */
    if ( ( ( sym.n_type ) & N_TYPE ) != N_SECT )
    {
      continue;
    }

  /* Anexo adelante del nombre de la funcion la libreria a la cual pertenece */
    if ( strrchr ( path , '/' ) != NULL )
    {
      function_name  = strrchr ( path , '/' ) + 1;
      function_name += "!";
      function_name += ( function_names + sym.n_un.n_strx );
    }
    else
    {
      function_name  = path;
      function_name += "!";
      function_name += ( function_names + sym.n_un.n_strx );
    }

  /* Agrego un simbolo mas a la lista */
//    printf ( "%.8x %s\n" , sym.n_value  , function_names + sym.n_un.n_strx );
    this -> address_to_name [ sym.n_value ] = function_name;
    this -> name_to_address [ function_name ] = sym.n_value;
  }

/* Libero la zona de nombres de la memoria */
  free ( function_names );

/* Cierro el archivo */
  fclose ( f );

/* Salgo */
  return ( ret );
}

/****************************************************************************/
/****************************************************************************/

Debug::Debug ()
{
/* Seteos iniciales */
  this -> inferior_thread_count = 0;
  this -> inferior_threads = NULL;
  this -> tracear = FALSE;
  this -> direccion_breakpoint_a_recuperar = 0;
}

/****************************************************************************/

int Debug::Execute ( char *process_name , char **allargs )
{
  FILE *executable_file;
  char instruccion_original [ sizeof ( while_true_instruction ) ];
  char process_name_with_pointslash [ 512 + 1 ];
  unsigned int pos_in_file;
  unsigned int pos_in_memory;
  int child_pid;
  int ret = TRUE;
  int err;

/* Averiguo la direccion del ENTRY POINT */
  err = get_file_entry_point ( process_name , &pos_in_file , &pos_in_memory );

/* Si no pude encontrar el ENTRY POINT */
  if ( err == FALSE )
  {
    return ( FALSE );
  }

/* Ignoro el CTRL+C */
  signal ( SIGINT , SIG_IGN );

/* Abro el archivo para patchearlo */
  executable_file = fopen ( process_name , "r+b" );

/* Si el file no existe */
  if ( executable_file == NULL )
  {
    return ( FALSE );
  }

/* Me posiciono en el ENTRY POINT del file */
  fseek ( executable_file , pos_in_file , SEEK_SET );
  
/* Leo la primera instruccion */
  fread ( instruccion_original , sizeof ( while_true_instruction ) , 1 , executable_file );

/* Patcheo el file */
  fseek ( executable_file , pos_in_file , SEEK_SET );
  fwrite ( while_true_instruction , sizeof ( while_true_instruction ) , 1 , executable_file );

/* Cierro el file */
  fclose ( executable_file );

/* Hago una copia del proceso */
  child_pid = fork ();  

/* Si soy el proceso hijo me transformo en el debugger */
  if ( child_pid == 0 )
  {
  /* Le doy tiempo al programa a que llegue al while(1) */
    sleep ( 1 );

  /* Me atacho al proceso padre */
    ret = this -> Attach ( getppid () );

  /* Si me puede attachar */
    if ( ret == TRUE )
    {
    /* Mientras el proceso no se haya cargado */
      while ( this -> GetEventCode () != DEBUG_EXCEPTION_CREATE_PROCESS )
      {
      /* Vuelvo a continuar el proceso */
        this -> Continue ();
        this -> Wait ();
      }

    /* Restauro la memoria */
      this -> WriteMemory ( ( void * ) pos_in_memory , ( void * ) instruccion_original , sizeof ( while_true_instruction ) );
    }
    else
    {
    /* Retorno ERROR */
      ret = FALSE;
    }

  /* Abro el archivo para restaurarlo */
    executable_file = fopen ( process_name , "r+b" );

  /* Me posiciono en el ENTRY POINT del file */
    fseek ( executable_file , pos_in_file , SEEK_SET );
  
  /* Patcheo el file con un while(1) para que me de tiempo a attacharme */
    fwrite ( &instruccion_original , sizeof ( while_true_instruction ) , 1 , executable_file );

  /* Cierro el file */
    fclose ( executable_file );
  }
/* Si soy el proceso padre ejecuto el programa pasado como parametro */
  else
  {
  /* Ejecuto el programa */
    err = execvp ( process_name , allargs );

  /* Si no pudo ejecutar el programa */
    if ( err == -1 )
    {
    /* Armo el nombre del ejecutable anexandole punto-barra al inicio */
    /* Esto pasa porque la funcion 'execvp' necesita el punto-barra si el file esta en el mismo directorio */
      strcpy ( process_name_with_pointslash , "./" );
      strncat ( process_name_with_pointslash , process_name , 512 );

    /* Intento ejecutarlo de vuelta */
     err = execvp ( process_name_with_pointslash , allargs );
    }

  /* Cuando finalize salgo */
    exit ( 0 );
  }

  return ( ret );
}

/****************************************************************************/

int Debug::Attach ( int pid )
{
  int ret = TRUE;
  int status;
  int err;
  int cont;

/* Catcheo la signal de CTRL+C para el debugger */
  signal ( SIGINT , ctrl_c );

/* Me guardo el pid */
  this -> pid = pid;

/* Si no puedo abrir el proceso */
  if ( task_for_pid ( mach_task_self () , pid , &this -> port ) != KERN_SUCCESS )
  {
    return ( FALSE );
  }

/* Suspendo el proceso */
  if ( task_suspend ( this -> port ) != KERN_SUCCESS )
  {
    return ( FALSE );
  }

/* Levanto la lista de threads */
  if ( task_threads ( this -> port , &this -> inferior_threads , &this -> inferior_thread_count ) != KERN_SUCCESS )
  {
    return ( FALSE );
  }

  if ( mach_port_allocate ( mach_task_self () , MACH_PORT_RIGHT_RECEIVE , &exception_port ) != KERN_SUCCESS )
  {
    fprintf(stderr, "Failed to create exception port.\n");
    return ( FALSE );
  }

  if ( mach_port_insert_right ( mach_task_self () , exception_port , exception_port , MACH_MSG_TYPE_MAKE_SEND ) != KERN_SUCCESS )
  {
    fprintf(stderr, "Failed to acquire insertion rights on the port.\n");
    return ( FALSE );
  }

  if ( task_set_exception_ports ( this -> port , EXC_MASK_ALL , exception_port , EXCEPTION_DEFAULT, THREAD_STATE_NONE ) != KERN_SUCCESS )
  {
    fprintf(stderr, "Failed to set the inferior's exception ports.\n");
    return ( FALSE );
  }

/* Seteo al thread cero como el que genero la excepcion */
  exception_thread = this -> inferior_threads [ 0 ];

/*
  printf ( "cantidad de threads = %i\n" , this -> inferior_thread_count );

  for ( cont = 0 ; cont < this -> inferior_thread_count ; cont ++ )
  {
    printf ( "thread %i\n" , inferior_threads [ cont ] );
  }
*/

  return ( ret );
}

/****************************************************************************/

int Debug::LoadSymbols ( char *filename )
{
  int ret;

/* Cargo los simbolos del proceso principal */
  ret = this -> load_symbol_table ( filename );

  printf ( "loaded symbols = %i\n" , this -> name_to_address.size () );

  return ( ret );
}

/****************************************************************************/

int Debug::Continue ( void )
{
  unsigned int addr_next_instruction;
  Breakpoint *breakpoint;

/* Si tengo que volver a habilitar el breakpoint */
  if ( this -> tracear == TRUE )
  {
  /* Para que no vuelva a entrar */
    this -> tracear = FALSE;

  /* Si el breakpoint no fue borrado */
    if ( this -> breakpoint_list.find ( this -> direccion_breakpoint_a_recuperar ) != this -> breakpoint_list.end () )
    {
    /* Calculo la direccion de la instruccion siguiente */
      addr_next_instruction = this -> direccion_breakpoint_a_recuperar + this -> len_breakpoint_a_recuperar;

    /* Si no hay un breakpoint seteado en la proxima instruccion */
      if ( this -> breakpoint_list.find ( addr_next_instruction ) == this -> breakpoint_list.end () )
      {
      /* Seteo un breakpoint en la proxima instruccion */
        this -> SetOneShotBreakpoint ( addr_next_instruction );

      /* Marco al breakpoint como recuperador */
        breakpoint = this -> breakpoint_list [ addr_next_instruction ];
        breakpoint -> recuperador = TRUE;
        breakpoint -> system = TRUE;
        breakpoint -> direccion_breakpoint_a_recuperar = this -> direccion_breakpoint_a_recuperar;
      }
    /* Si ya hay un breakpoint seteado */
      else
      {
      /* Marco al breakpoint como recuperador */
        breakpoint = this -> breakpoint_list [ addr_next_instruction ];
        breakpoint -> recuperador = 1;
        breakpoint -> direccion_breakpoint_a_recuperar = this -> direccion_breakpoint_a_recuperar;
      }
    }
  }

/* Aca deberia flushearse el cache del procesador */
  sleep ( 1 );

/* Continuo la ejecucion del proceso */
  task_resume ( this -> port );

  return ( TRUE );
}

/****************************************************************************/

int Debug::Wait ( void )
{
  Breakpoint *breakpoint;
  int breakpoint_recuperador;
  int breakpoint_system;
  int msg [ 128 ];
  int out_msg [ 128 ];
  int len_msg;
  int cont;
  int ret = TRUE;
  int err;

/* El proceso esta corriendo, esto es para el CTRL+C */
  running = TRUE;

/* Espero una excepcion del proceso */
  while ( ( ( err = mach_msg ( ( mach_msg_header_t * ) msg , MACH_RCV_MSG | MACH_RCV_TIMEOUT , 0 , sizeof ( msg ) , this -> exception_port , 1000 , MACH_PORT_NULL ) ) == MACH_RCV_TIMED_OUT ) && ( breakear == FALSE ) )
  {
  /* Intento suspender el proceso para ver si todavia esta corriendo */
    err = task_suspend ( this -> port );

  /* Si no lo pude suspender */
    if ( err != KERN_SUCCESS )
    {
    /* Retorno una excepcion de fin de proceso */
      exception_type = DEBUG_EXCEPTION_EXIT_PROCESS;
      
    /* No puedo debuggear mas el proceso */
      return ( TRUE );
    }
    else
    {
    /* Lo vuelvo a activar */
      task_resume ( this -> port );
    }
  }

/* Reseteo el flag */
  running = FALSE;

/* Si se produjo un breakeo */
  if ( breakear == TRUE )
  {
  /* Retorno OK */
    err = KERN_SUCCESS;
  }

/* Si es alguna excepcion rara */
  if ( err != KERN_SUCCESS )
  {
    return ( FALSE );
  }

/* Suspendo el proceso */
  task_suspend ( this -> port );

/* Si se produjo un break salteo la parte de abajo */
  if ( breakear == TRUE )
  {
  /* Lo reseteo */
    breakear = FALSE;

  /* Retorno OK */
    return ( TRUE );
  }

/* Mando un mensaje a mi mismo para saber que excepcion es */
  exc_server( ( mach_msg_header_t * ) msg , ( mach_msg_header_t * ) out_msg );

/* Longitud del mensaje a enviar */
  len_msg = ( unsigned ) out_msg [ 1 ];

/* Envio una exception reply */
  if ( mach_msg ( ( mach_msg_header_t * ) out_msg , MACH_SEND_MSG , len_msg , 0 , MACH_PORT_NULL , 0 , MACH_PORT_NULL ) != KERN_SUCCESS )
  {
  /* Retorno ERROR */
    return ( FALSE );
  }

/* Actualizo la lista de threads */
  task_threads ( this -> port , &this -> inferior_threads , &this -> inferior_thread_count );

/* Si se produjo un breakpoint */
  if ( this -> GetEventCode () == DEBUG_EXCEPTION_BREAKPOINT )
  {
  /* Si el breakpoint es mio */
    if ( this -> breakpoint_list.find ( this -> GetExceptionAddress () ) != this -> breakpoint_list.end () )
    {
    /* Leo el breakpoint */
      breakpoint = this -> breakpoint_list [ this -> GetExceptionAddress () ];

    /* Averiguo si este breakpoint tiene que recuperar a otro */
      breakpoint_recuperador = breakpoint -> recuperador;

    /* Averiguo si es un breakpoint de sistema */
      breakpoint_system = breakpoint -> system;

    /* Si el breakpoint no es oneshot */
      if ( breakpoint -> oneshot == FALSE )
      {
      /* Si tengo que recuperar un breakpoint */
        if ( breakpoint_recuperador != 0 )
        {
        /* Recupero el breakpoint */
          this -> EnableBreakpoint ( breakpoint -> direccion_breakpoint_a_recuperar );

        /* Marco al breakpoint para que no recupere de vuelta */
          breakpoint -> recuperador = 0;
        }

      /* Cuando continue con el proceso tengo que tracear una instruccion */
        this -> tracear = TRUE;

      /* Deshabilito el breakpoint */
        this -> DisableBreakpoint ( breakpoint -> address );

      /* Direccion donde ocurrio el breakpoint original */
        this -> direccion_breakpoint_a_recuperar = this -> GetExceptionAddress ();

      /* Para calcular donde esta la siguiente instruccion */
        this -> len_breakpoint_a_recuperar = breakpoint -> len_instruccion_original;
      }
    /* Si el breakpoint es oneshot */
      else
      {
      /* Si tengo que recuperar un breakpoint anterior */
        if ( breakpoint_recuperador != 0 )
        {
        /* Recupero el breakpoint */
          this -> EnableBreakpoint ( breakpoint -> direccion_breakpoint_a_recuperar );

        /* Elimino el breakpoint */
          this -> UnsetBreakpoint ( breakpoint -> address );

        /* Si este breakpoint lo seteo la misma clase para recuperar uno anterior */
          if ( ( breakpoint_recuperador == TRUE ) && ( breakpoint_system == TRUE ) )
          {
          /* Absorbo la excepcion para que el usuario no la vea y pueda restaurar el breakpoint */
            this -> Continue ();
            this -> Wait ();
          }
        }
        else
        {
        /* Elimino el breakpoint */
          this -> UnsetBreakpoint ( breakpoint -> address );
        }
      }
    }
  }

  return ( ret );
}

/****************************************************************************/

int Debug::Close ( void )
{
  int ret = FALSE;

/* Cierro el proceso debuggeado */
  if ( kill ( this -> pid , SIGQUIT ) == 0 )
  {
  /* Retorno OK */
    ret = TRUE;
  }

  return ( ret );
}

/****************************************************************************/

int Debug::GetPc ( void )
{
  unsigned int regs [ CANTIDAD_REGISTROS ];
  int ret;

/* Leo los registros */
  this -> GetRegisters ( this -> GetCurrentTid () , regs );

/* Retorno el valor del registro PC */
  ret = regs [ REGISTER_PC ];

  return ( ret );
}

/****************************************************************************/

int Debug::GetCurrentPid ( void )
{
/* Retorno el proceso que genero la excepcion */
  return ( pid );
}

/****************************************************************************/

int Debug::GetCurrentTid ( void )
{
/* Retorno el thread que genero la excepcion */
  return ( exception_thread );
}

/****************************************************************************/

int *Debug::GetThreadList ( int *cantidad )
{
/* Retorno la cantidad actual de threads */
  *cantidad = this -> inferior_thread_count;

/* Retorno la lista de threads */
  return ( ( int * ) this -> inferior_threads );
}

/****************************************************************************/

// Para que ponga un breakpoint en modo THUMB hay que pasarle "address + 1"
// Es para que el metodo sepa distinguir entre un modo y el otro

int Debug::SetBreakpoint ( unsigned int address )
{
  Breakpoint *breakpoint;
  char instruccion_original [ 4 ];
  unsigned int bytes_leidos;
  unsigned int len_breakpoint;
  int breakpoint_mode;
  int ret = FALSE;

/* Si la direccion esta en modo THUMB */  
  if ( address & 0x01 )
  {
  /* Logitud y tipo del breakpoint segun el modo */
    len_breakpoint = 2;
    breakpoint_mode = 0;
  }
  else
  {
  /* Logitud y tipo del breakpoint segun el modo */
    len_breakpoint = 4;
    breakpoint_mode = 1;
  }

/* Si no hay un breakpoint es esta direccion */
  if ( this -> breakpoint_list.find ( address ) == this -> breakpoint_list.end () )
  {
  /* Guardo la instruccion original */
    if ( this -> ReadMemory ( ( void * ) instruccion_original , ( void * ) ( address & 0xfffffffe ) , len_breakpoint , &bytes_leidos ) == TRUE )
    {
    /* Escribo un breakpoint en esa direccion */
      ret = this -> WriteMemory ( ( void * ) ( address & 0xfffffffe ) , ( void * ) breakpoint_instruction [ breakpoint_mode ] , len_breakpoint );

    /* Si pude escribir el breakpoint */
      if ( ret == TRUE )
      {
      /* Creo una estructura para el breakpoint */
        breakpoint = ( Breakpoint * ) malloc ( sizeof ( Breakpoint ) );  

      /* Seteo las propiedades del breakpoint */
        breakpoint -> enabled = TRUE;
        breakpoint -> oneshot = FALSE;
        breakpoint -> recuperador = 0;
        breakpoint -> system = FALSE;
        breakpoint -> address = address;
        memcpy ( breakpoint -> instruccion_original , instruccion_original , len_breakpoint );
        breakpoint -> len_instruccion_original = len_breakpoint;
        breakpoint -> breakpoint_mode = breakpoint_mode;

      /* Agrego el breakpoint a la lista */
        this -> breakpoint_list [ address ] = breakpoint;
      }
    }
  }

  return ( ret );
}

/****************************************************************************/

int Debug::UnsetBreakpoint ( unsigned int address )
{
  Breakpoint *breakpoint;
  int ret = FALSE;

/* Si hay un breakpoint es esta direccion */
  if ( this -> breakpoint_list.find ( address ) != this -> breakpoint_list.end () )
  {
  /* Leo el breakpoint de la lista */
    breakpoint = this -> breakpoint_list [ address ];

  /* Restauro el breakpoint */
    ret = this -> WriteMemory ( ( void * ) ( address & 0xfffffffe ) , ( void * ) breakpoint -> instruccion_original , breakpoint -> len_instruccion_original );  

  /* Elimino el breakpoint de la lista */
    this -> breakpoint_list.erase ( address );

  /* Elimino la estructura del breakpoint */
    free ( breakpoint );
  }

  return ( ret );
}

/****************************************************************************/

int Debug::SetOneShotBreakpoint ( unsigned int address )
{
  Breakpoint *breakpoint;
  int ret;

/* Seteo el breakpoint */  
  ret = this -> SetBreakpoint ( address );

/* Si pude setear el breakpoint */
  if ( ret == TRUE )
  {
  /* Seteo la propiedad oneshot en la estructura del breakpoint */
    breakpoint = this -> breakpoint_list [ address ];
    breakpoint -> oneshot = TRUE;
  }

  return ( ret );
}

/****************************************************************************/

int Debug::EnableBreakpoint ( unsigned int address )
{
  Breakpoint *breakpoint;
  int ret = FALSE;

/* Si hay un breakpoint es esta direccion */
  if ( this -> breakpoint_list.find ( address ) != this -> breakpoint_list.end () )
  {
  /* Leo el breakpoint de la lista */
    breakpoint = this -> breakpoint_list [ address ];

  /* Si el breakpoint esta deshabilitado */
    if ( breakpoint -> enabled == FALSE )
    {
    /* Lo marco como habilitado */
      breakpoint -> enabled = TRUE;

    /* Restauro el breakpoint */
      ret = this -> WriteMemory ( ( void * ) ( address & 0xfffffffe ) , ( void * ) breakpoint_instruction [ breakpoint -> breakpoint_mode ] , breakpoint -> len_instruccion_original );  

    /* Retorno OK */
      ret = TRUE;
    }
  }

  return ( ret );
}

/****************************************************************************/

int Debug::DisableBreakpoint ( unsigned int address )
{
  Breakpoint *breakpoint;
  int ret = FALSE;

/* Si hay un breakpoint es esta direccion */
  if ( this -> breakpoint_list.find ( address ) != this -> breakpoint_list.end () )
  {
  /* Leo el breakpoint de la lista */
    breakpoint = this -> breakpoint_list [ address ];

  /* Si el breakpoint esta habilitado */
    if ( breakpoint -> enabled == TRUE )
    {
    /* Lo marco como deshabilitado */
      breakpoint -> enabled = FALSE;

    /* Restauro la instruccion original */
      ret = this -> WriteMemory ( ( void * ) ( address & 0xfffffffe ), ( void * ) breakpoint -> instruccion_original , breakpoint -> len_instruccion_original );  

    /* Retorno OK */
      ret = TRUE;
    }
  }

  return ( ret );
}

/****************************************************************************/

list <BreakpointInfo *> *Debug::GetBreakpointList ( void )
{
  map <unsigned int,Breakpoint *> :: iterator it_breakpoint;
  list <BreakpointInfo *> :: iterator it_user_breakpoint;
  BreakpointInfo *user_breakpoint;
  Breakpoint *breakpoint;

/* Recorro toda la lista borrando las estructuras anteriores */
  for ( it_user_breakpoint = this -> user_breakpoint_list.begin () ; it_user_breakpoint != this -> user_breakpoint_list.end () ; it_user_breakpoint ++ )
  {
  /* Libero la estructura */
    free ( *it_user_breakpoint );

  /* Elimino el breakpoint de la lista */
    user_breakpoint_list.erase ( it_user_breakpoint );
  }

/* Recorro toda la lista real de breakpoints */
  for ( it_breakpoint = this -> breakpoint_list.begin () ; it_breakpoint != this -> breakpoint_list.end () ; it_breakpoint ++ )
  {
  /* Leo el siguiente breakpoint */
    breakpoint = ( Breakpoint * ) it_breakpoint -> second;

  /* Si el breakpoint no lo puso la misma clase */
    if ( breakpoint -> system == FALSE )
    {
    /* Agrego un breakpoint mas para mostrar */
      user_breakpoint = ( BreakpointInfo * ) malloc ( sizeof ( BreakpointInfo ) );

    /* Propiedades que me interesan */
      user_breakpoint -> address = breakpoint -> address;
      user_breakpoint -> enabled = breakpoint -> enabled;
      user_breakpoint -> oneshot = breakpoint -> oneshot;

    /* Agrego el breakpoint a la lista */
      user_breakpoint_list.push_back ( user_breakpoint );
    }
  }

  return ( &user_breakpoint_list );
}

/****************************************************************************/

int Debug::EnableTrace ( void )
{
  return ( 0 );
}

/****************************************************************************/

int Debug::DisableTrace ( void )
{
  return ( 0 );
}

/****************************************************************************/

int Debug::ReadMemory ( void *dest , void *source , unsigned int len , unsigned int *leidos )
{
  int ret = TRUE;

/* Si no pude leer la memoria del proceso */
  if ( vm_read_overwrite ( this -> port , ( unsigned int ) source , len , ( vm_address_t ) dest , leidos ) != 0 )
  {
    ret = FALSE;
  }

  return ( ret );
}

/****************************************************************************/

int Debug::WriteMemory ( void *dest , void *source , unsigned int len )
{
  unsigned int attributes;
  struct vm_region_basic_info info;
  unsigned int base_dest = ( unsigned int ) dest;
  int len_info = sizeof ( struct vm_region_basic_info );
  int len_region = len;
  int ret = TRUE;
  int obj_name;
  int err;

/* Pongo la zona de memoria como READ-WRITE */
  attributes = DEBUG_PROTECTION_READ | DEBUG_PROTECTION_WRITE;

/* Cambio los permisos de la zona para poder escribir la memoria */
  if ( this -> SetMemoryAttributes ( ( unsigned int ) dest , len , &attributes ) == FALSE )
  {
  /* Retorno ERROR */
    return ( FALSE );
  }

/* Si no pude escribir la memoria del proceso */
  if ( vm_write ( this -> port , ( vm_address_t ) dest , ( vm_offset_t ) source , len ) != 0 )
  {
  /* Retorno ERROR */
    ret = FALSE;
  }

/* Restauro los permisos de la zona */
  if ( this -> SetMemoryAttributes ( ( unsigned int ) dest , len , &attributes ) == FALSE )
  {
  /* Retorno ERROR */
    ret = FALSE;
  }

  return ( ret );
}

/****************************************************************************/

int Debug::Search ( void *addr_base , void *addr_limite , char *pattern , unsigned int len , unsigned int *addr_found )
{
  char buffer [ 4096 * 2 ];
  unsigned int paginas_leidas;
  unsigned int bytes_leidos;
  unsigned int addr_alineada;
  unsigned int addr_offset;
  unsigned int offset;
  unsigned int pos;
  unsigned int addr;
  int ret = FALSE;

/* Recorro toda la memoria buscando el patron */
  for ( addr = ( unsigned int ) addr_base ; ( addr < ( unsigned int ) addr_limite ) && ( addr < 0xfffff000 ) ; addr = addr + offset )
  {
  /* Variables para no repetir operaciones aritmeticas */
    addr_alineada = ( addr / 4096 ) * 4096;
    addr_offset = addr % 4096;

  /* Si estoy en el medio de una pagina */
    if ( addr_offset != 0 )
    {
    /* Corrijo el offset para alinearme a la pagina */
      offset = 4096 - addr_offset;
    }
    else
    {
    /* Me muevo de a 1 pagina */
      offset = 4096;
    }

  /* Reseteo el contador de paginas leidas */
    paginas_leidas = 0;

  /* Levanto la pagina donde estoy parado */
    if ( this -> ReadMemory ( buffer , ( void * ) addr_alineada , 4096 , &bytes_leidos ) == TRUE )
    {
    /* La pagina actual */
      paginas_leidas ++;

    /* Si puedo leer la proxima pagina */
      if ( this -> ReadMemory ( &buffer [ 4096 ] , ( void * ) ( addr_alineada + 4096 ) , 4096 , &bytes_leidos ) == TRUE )
      {
      /* La pagina sigiente */
        paginas_leidas ++;
      }
    }
  /* Si no pude leer la pagina avanzo a la proxima */
    else
    {
    /* Sigo buscando */
      continue;
    }

  /* Busco dentro de la pagina */
    for ( pos = addr_offset ; pos < 4096 ; pos ++ )
    {
    /* Si llegue al limite */
      if ( ( addr_alineada + pos + len - 1 ) >= ( unsigned int ) addr_limite )
      {
      /* Salgo */
        return ( FALSE );
      }

    /* Si la longitud del patron sobrepasa la pagina */
      if ( ( pos + len ) >= 4096 )
      {
      /* Si no tengo una segunda pagina */
        if ( paginas_leidas < 2 )
        {
        /* Sigo cargando paginas */
          break;
        }
      }

    /* Si es el patron que estoy buscando */
      if ( memcmp ( &buffer [ pos ], pattern , len ) == 0 )
      {
      /* Retorno la direccion donde encontre el patron */
        *addr_found = addr_alineada + pos;
  
      /* Retorno OK */
        return ( TRUE );
      }
    }
  }

  return ( ret );
}

/****************************************************************************/

int Debug::GetRegisters ( int thread , unsigned int *registers )
{
  unsigned int cant_registros = CANTIDAD_REGISTROS;
  int ret = FALSE;

/* Si pude leer el estado del thread pasado como parametro */
  if ( ( thread_get_state ( thread , TRUE , ( thread_state_t ) registers , &cant_registros ) ) == KERN_SUCCESS )
  {
  /* Si esta en modo THUMB modifico el PC */
    if ( registers [ REGISTER_CTRL ] & MODE_THUMB_ENABLED )
    {
    /* Prendo el bit cero del registro PC */
      registers [ REGISTER_PC ] = registers [ REGISTER_PC ] | 0x01;
    }

  /* Retorno OK */
    ret = TRUE;
  }

  return ( ret );
}

/****************************************************************************/

int Debug::SetRegisters ( int thread , unsigned int *registers )
{
  unsigned int cant_registros = CANTIDAD_REGISTROS;
  int ret = TRUE;

/* Seteo el estado del thread 0 */
  if ( ( thread_set_state ( thread , TRUE , ( thread_state_t ) registers , ( mach_msg_type_number_t ) &cant_registros ) ) != KERN_SUCCESS )
  {
  /* Retorno ERROR */
    ret = FALSE;
  }

  return ( ret );
}

/****************************************************************************/

int Debug::GetEventCode ( void )
{
/* Retorno el codigo de excepcion */
  return ( exception_type );
}

/****************************************************************************/

int Debug::GetExceptionAddress ( void )
{
/* Retorno la direccion donde se produjo la excepcion */
  return ( this -> GetPc () );
}

/****************************************************************************/

int Debug::ResolveAddressToName ( unsigned int address , char *name , int *offset )
{
  map <unsigned int,string> :: iterator symbol;
  unsigned int address_cercana = 0;
  int ret = TRUE;

/* Mientras no llegue al final de la lista */
  for ( symbol = this -> address_to_name.begin () ; symbol != this -> address_to_name.end () ; symbol ++ )
  {
  /* Si la direccion es mas cercana que la anterior */
    if ( ( ( *symbol ).first <= address ) && ( address_cercana < ( *symbol ).first ) )
    {
    /* Tengo una nueva address */
      address_cercana = ( *symbol ).first;

    /* Calculo el offset al nombre de la funcion */
      *offset = address - address_cercana;
    }
  }

/* Si pude encontrar una direccion */
  if ( address_cercana != 0 )
  {
  /* Copio el nombre al destino */
    strcpy ( name , this -> address_to_name [ address_cercana ].c_str () );
  }
  else
  {
    ret = FALSE;
  }

  return ( ret );
}

/****************************************************************************/

int Debug::ResolveNameToAddress ( char *name , unsigned int *address )
{
  map <string,unsigned int> :: iterator symbol;
  string symbol_name;
  int ret = FALSE;

/* Mientras no llegue al final de la lista */
  for ( symbol = this -> name_to_address.begin () ; symbol != this -> name_to_address.end () ; symbol ++ )
  {
  /* Nombre del proximo simbolo */
    symbol_name = ( *symbol ).first;

  /* Si encuentro el nombre dentro del simbolo */
    if ( strcmp ( ( char * ) ( strchr ( symbol_name.c_str () , '!' ) + 1 ) , name ) == 0 )
    {
    /* Direccion en la memoria del simbolo */
      *address = ( *symbol ).second;

    /* Retorno OK */
      ret = TRUE;    

    /* Corto la busqueda */
      break;
    }
  }

  return ( ret );
}

/****************************************************************************/

// Funciona igual que 'strtok', la primera vez se le pasa un string
// La segunda vez que es llamada se le pasa NULL para que retorne el proximo simbolo matcheado

int Debug::GetSymbolFrom ( char *lib_function , unsigned int *address , const char **psymbol )
{
  string symbol_name;
  char *separador;
  int ret = FALSE;

/* Si es la primera vez */
  if ( lib_function != NULL )
  {
  /* Separo el nombre de la lib y de la funcion */
    separador = strchr ( lib_function , '!' );

  /* Si el nombre del simbolo esta mal pasado */
    if ( separador == NULL )
    {
    /* Retorno ERROR */
      return ( FALSE );
    }

  /* Me ubico al principio de la lista de simbolos */
    this -> symbol = this -> name_to_address.begin ();

  /* Me guardo el nombre de la lib */
    snprintf ( this -> libname , 256 , "%s" , lib_function , separador - lib_function );

  /* Si quiero buscar en todas las libs */
    if ( strncmp ( this -> libname , "*" , 1 ) == 0 )
    {
    /* Marco el flag */
      this -> especific_lib = FALSE;
    }
  /* Si quiero buscar en alguna lib especifica */
    else
    {
    /* Marco el flag */
      this -> especific_lib = TRUE;
    }

  /* Me quedo con el nombre de la funcion */
    snprintf ( this -> funcname , 256 , "%s" , separador + 1 );

  /* Si quiero buscar en todas las funciones */
    if ( strncmp ( this -> funcname , "*" , 1 ) == 0 )
    {
    /* Marco el flag */
      this -> especific_function = FALSE;
    }
  /* Si quiero buscar en alguna funcion especifica */
    else
    {
    /* Marco el flag */
      this -> especific_function = TRUE;
    }
  }

/* Me traigo las variables de la clase */

/* Recorro la lista de todos los simbolos cargados */
  for ( ; this -> symbol != this -> name_to_address.end () ; this -> symbol ++ )
  {
  /* Nombre del proximo simbolo */
    symbol_name = ( *symbol ).first;

  /* Busco el separador entre lib y funcion */
    separador = strchr ( symbol_name.c_str () , '!' );

  /* Si tengo que buscar el nombre de una lib en particular */
    if ( this -> especific_lib == TRUE )
    {
    /* Si no es la lib que estoy buscando */
      if ( strncmp ( this -> libname , symbol_name.c_str () , separador - symbol_name.c_str () ) != 0 )
      {
      /* Sigo buscando */
        continue;
      }
    }

  /* Si tengo que buscar el nombre de una funcion en particular */
    if ( this -> especific_function == TRUE )
    {
    /* Si no es la funcion que estoy buscando */
      if ( strcmp ( this -> funcname , separador + 1 ) != 0 )
      {
      /* Sigo buscando */
        continue;
      }
    }

  /* Si llegue hasta aca es porque cumple todas las condiciones del simbolo a buscar */
  /* Retorno la direccion y el nombre del simbolo encontrado */
    *address = ( *symbol ).second;
    *psymbol = symbol_name.c_str ();

  /* Avanzo una posicion en la lista */
    this -> symbol ++;

  /* Retorno OK */
    ret = TRUE;

  /* Corto la busqueda */
    break;
  }

  return ( ret );
}

/****************************************************************************/

int Debug::GetMemoryAttributes ( unsigned int address , unsigned int *base , unsigned int *len , unsigned int *attributes )
{
  struct vm_region_basic_info info;
  unsigned int base_dest = address;
  unsigned int leidos;
  int len_info = sizeof ( struct vm_region_basic_info );
  int len_region;
  int ret = FALSE;
  int obj_name;
  int err;
  char c;

/* Leo los permisos de la zona de memoria */
  err = vm_region ( this -> port , ( vm_address_t * ) &base_dest , ( vm_address_t * ) &len_region , VM_REGION_BASIC_INFO , ( vm_region_info_t ) &info , ( mach_msg_type_number_t * ) &len_info , ( mach_port_t * ) &obj_name );

/* Si hay memoria mapeada en esa zona y la puedo leer */
  if ( ( err != KERN_INVALID_ADDRESS ) && ( this -> ReadMemory ( &c , ( void * ) base_dest , 1 , &leidos ) == TRUE ) )
  {
  /* Retorno los permisos de la zona */
    *base = base_dest;
    *len = len_region;
    *attributes = 0;

   /* Si la zona tiene el atributo de lectura */ 
    if ( info.protection & VM_PROT_READ )
    {
      *attributes = *attributes | DEBUG_PROTECTION_READ;
    }

   /* Si la zona tiene el atributo de escritura */ 
    if ( info.protection & VM_PROT_WRITE )
    {
      *attributes = *attributes | DEBUG_PROTECTION_WRITE;
    }

   /* Si la zona tiene el atributo de ejecucion */ 
    if ( info.protection & VM_PROT_EXECUTE )
    {
      *attributes = *attributes | DEBUG_PROTECTION_EXECUTE;
    }

  /* Retorno OK */
    ret = TRUE;
  }

  return ( ret );
}

/****************************************************************************/

int Debug::SetMemoryAttributes ( unsigned int address , unsigned int len , unsigned int *attributes )
{
  struct vm_region_basic_info info;
  unsigned int attributes_to_set;
  unsigned int base_dest = address;
  int len_info = sizeof ( struct vm_region_basic_info );
  int len_region = len;
  int obj_name;
  int ret = FALSE;
  int err;

/* Leo los permisos de la zona de memoria */
  err = vm_region ( this -> port , ( vm_address_t * ) &base_dest , ( vm_address_t * ) &len_region , VM_REGION_BASIC_INFO , ( vm_region_info_t ) &info , ( mach_msg_type_number_t * ) &len_info , ( mach_port_t * ) &obj_name );

/* Si pude leer los permisos de la zona */
  if ( err == KERN_SUCCESS )
  {
  /* Inicializo los atributos a setear */
    attributes_to_set = 0;

  /* Si la zona tiene el atributo de lectura */
    if ( *attributes & DEBUG_PROTECTION_READ )
    {
      attributes_to_set = attributes_to_set | VM_PROT_READ ;
    }

  /* Si la zona tiene el atributo de escritura */ 
    if ( *attributes & DEBUG_PROTECTION_WRITE )
    {
      attributes_to_set = attributes_to_set | VM_PROT_WRITE;
    }

  /* Si la zona tiene el atributo de ejecucion */ 
    if ( *attributes & DEBUG_PROTECTION_EXECUTE )
    {
      attributes_to_set = attributes_to_set | VM_PROT_EXECUTE;
    }

  /* Cambio los permisos de la zona */
    if ( vm_protect ( this -> port , ( vm_address_t ) base_dest , len_region , FALSE , attributes_to_set ) == KERN_SUCCESS )
    {
    /* Inicializo los atributos a retornar */
      *attributes = 0;

    /* Si la zona tiene el atributo de lectura */ 
      if ( info.protection & VM_PROT_READ )
      {
        *attributes = *attributes | DEBUG_PROTECTION_READ;
      }

     /* Si la zona tiene el atributo de escritura */ 
      if ( info.protection & VM_PROT_WRITE )
      {
        *attributes = *attributes | DEBUG_PROTECTION_WRITE;
      }

     /* Si la zona tiene el atributo de ejecucion */ 
      if ( info.protection & VM_PROT_EXECUTE )
      {
        *attributes = *attributes | DEBUG_PROTECTION_EXECUTE;
      }

    /* Retorno OK */
      ret = TRUE;
    }
  }

  return ( ret );
}

/****************************************************************************/

void print_address_func ( bfd_vma v1 , disassemble_info *v2 )
{
  char buffer [ 256 ];

/* Transformo la direccion a saltar en un branch */
  sprintf ( buffer , "%x" , v1 );

/* Anexo el branch a la instruccion */
//  strcat ( traduccion , buffer );
  strcat ( ( char * ) v2 -> buffer , buffer );
}

/****************************************************************************/

int my_fprintf ( FILE *_stream , const char *format , ... )
{
  char buffer [ 256 ];
  void *arg [ 5 ];
  void *actual_arg;
  va_list args;
  int cont;

/* Leo el primer argumento */
  va_start ( args , format );  
  actual_arg = ( void * ) format;

/* Recorro todos los argumentos */
  for ( cont = 0 ; ( actual_arg != NULL ) && ( cont < 5 ) ; cont ++ )
  {
  /* Leo el proximo argumento */  
    actual_arg = va_arg ( args , void * );
    arg [ cont ] = actual_arg;
  }

/* Reteo los argumentos */
  va_end ( args );

/* Escribo la instruccion en mi buffer */
  sprintf ( buffer , format , arg [ 0 ] , arg [ 1 ] , arg [ 2 ] , arg [ 3 ] , arg [ 4 ] , arg [ 5 ] );

/* Voy armando la instruccion con los reiterados llamados a la funcion */
  strcat ( ( char * ) _stream , buffer );
}

/****************************************************************************/

int Debug::Disassembly ( unsigned int address , unsigned int *len , char *dest_buffer )
{
  unsigned char opcodes [ MAX_LEN_INSTRUCTION ];
  struct disassemble_info di;
  unsigned int bytes_leidos;
  int ret = TRUE;
  int err;

/* Levanto una parte de la zona de memoria a desensamblar */
  err = this -> ReadMemory ( ( void * ) opcodes , ( void * ) ( address & 0xfffffffe ) , MAX_LEN_INSTRUCTION , &bytes_leidos );

/* Si no pude leer a memoria */
  if ( err == FALSE )
  {
  /* Retorno ERROR */
    return ( FALSE );
  }

/* Seteo las opciones de la libopcode */
  di.flavour = bfd_target_unknown_flavour;
  di.arch = bfd_arch_unknown;
  di.mach = 0;
  di.endian = BFD_ENDIAN_LITTLE;

/* Funcion que se encarga de simular un STREAM de salida */
  di.fprintf_func = my_fprintf;
  di.stream = ( FILE * ) dest_buffer;

  di.symbols = NULL;
  di.num_symbols = 0;

/* Buffer donde va a parar la instruccion desensamblada */
  di.buffer = ( bfd_byte * ) dest_buffer;
  di.buffer_vma = 0;
  di.buffer_length = 0;
  di.read_memory_func = NULL;
  di.memory_error_func = NULL;

/* Puntero a funcion que es llamada cuando */
/* hay una referencia a una direccion de memoria */
  di.print_address_func = print_address_func;

  di.symbol_at_address_func = NULL;
  di.flags = 0;
  di.bytes_per_line = 0;
  di.bytes_per_chunk = 0;
  di.display_endian = BFD_ENDIAN_UNKNOWN;
  di.disassembler_options = NULL;
  di.insn_info_valid = 0;

/* Valor por default en caso de error */
  di.bytes_per_chunk = 4;

/* Inicializo el buffer donde va a parar la instruccion */
  strcpy ( dest_buffer , "" );

/* Desensamblo la instruccion */
  print_insn_arm ( address , &di , opcodes );

/* Bytes usados por la instruccion */
  *len = di.bytes_per_chunk;

  return ( ret );
}

/****************************************************************************/
/****************************************************************************/
