/****************************************************************************/
/****************************************************************************/

/*
   Console interface for debugging.
   Copyright (c) 2008 Core Security Technologies, Core SDI Inc.
   Developed by Nicolas A. Economou (neconomou@corest.com).

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your option)
any later version. 

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
more details. 

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/****************************************************************************/
/****************************************************************************/

/* iphonedbg.c */

/****************************************************************************/
/****************************************************************************/

/* Definicion de las clases */

struct elemento
{
  char usado;
  void *elemento;
};

typedef struct
{
  unsigned int posicion_query;
  struct elemento *ultimo_query;
  unsigned int cantidad_elementos;
  unsigned int cantidad_maxima;
  struct elemento *elementos;
} Bag;

/****************************************************************************/
/****************************************************************************/

/* Includes */

#define FALSE 0
#define TRUE  1

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "iphone_debug.cpp"

/****************************************************************************/
/****************************************************************************/

/* Estructuras */

typedef struct
{
  unsigned int address;
  int conditional;
  char condition [ 32 + 1 ];
} _breakpoint;

/****************************************************************************/

/* Prototipos */

int continue2 ( Debug * );
int preprocess_command ( Debug * , char * );
int parsear_argumentos ( char * , char ** );

unsigned int get_register_value ( Debug * , char * );
int set_register_value ( Debug * , char * , char * );
unsigned int string_to_direccion ( Debug * , char * );
int registers_to_addresses ( Debug * , char * );
int names_to_addresses ( Debug * , char * );
int addresses_to_address ( char * );
unsigned int register_to_address ( Debug * , char * );
unsigned int selector_to_address ( Debug * , char * );
int is_register ( char * , unsigned int * );
int isCorrectSintaxCondition ( char * );
int isConditionalBreakpoint ( unsigned int );
int isConditionalBreakpointTrue ( Debug * );

unsigned int imprimir_memoria_byte ( Debug * , unsigned int , unsigned int );
unsigned int imprimir_memoria_dword ( Debug * , unsigned int , unsigned int );
unsigned int imprimir_codigo ( Debug * , unsigned int );
void ascii_bytes_to_bytes ( char * , char ** , int );
void imprimir_registros ( Debug * , int );

char *bin2asm ( Debug * , unsigned int , unsigned int * );
char *strins ( char * , char * );
char *strsup ( char * , int );

/****************************************************************************/

/* Defines */

#define VALUE     1
#define REGISTER  2
#define MEMORY    3

/****************************************************************************/

/* Variables Globales */

char *version = "iphonedbg v1.01";
unsigned int contador_steps = 1;

// Los operadores usados para calcular 'expression'
char *operadores = "-+";

// Diccionario con la lista de breakpoints
map <unsigned int,_breakpoint *> breakpoints;

/****************************************************************************/
/****************************************************************************/

/* Programa principal */

int main ( int argc , char *argv [] )
{
  Debug *proceso;
  char ** allargs;
  char buffer [ 256 ];
  char buffer2 [ 256 ];
  char *cmd;
  int attachado;
  int repeats;
  int cont;
  int pid;

/* Tomo los parametros */
  if ( argc < 3 )
  {
    printf ( "\n%s" , version );
    printf ( "\nCreated by Nicolas A. Economou ( neconomou@corest.com )" );
    printf ( "\nCore Security Technologies, Buenos Aires, Argentina ( 2008 )\n" );
    printf ( "\nsintax: iphonedbg <-e executable [arguments...]|-p pid executable>\n" );

    return ( 0 );
  }

/* Creo un objeto Debug */
  proceso = new ( Debug );

/* Si quiero ejecutar un proceso */
  if ( strcmp ( argv [ 1 ] , "-e" ) == 0 )
  {
  /* Si el comando tiene parametros pasados al ejecutable */
    if ( argc > 3 )
    {
      allargs = ( char ** ) &argv [ 3 ];
    }
    else
    {
      allargs = NULL;
    }

    if ( proceso -> Execute ( argv [ 2 ] , allargs ) == FALSE )
    {
      printf ( "\nprocess error !!!\n" );
      return ( 0 );
    }

  /* No estoy attachado */
    attachado = FALSE;

  /* Cargo los simbolos */
    puts ( "loading symbols ..." );
    proceso -> LoadSymbols ( argv [ 2 ] );
  }
  else if ( ( argc == 4 ) && ( strcmp ( argv [ 1 ] , "-p" ) == 0 ) )
  {
  /* Transformo el segundo argumento en un pid */
    pid = atoi ( argv [ 2 ] );

  /* Me atacho al proceso */
    if ( proceso -> Attach ( pid ) == FALSE )
    {
      printf ( "\nprocess error !!!\n" );
      return ( 0 );
    }

  /* Estoy attachado */
    attachado = TRUE;

  /* Cargo los simbolos */
    puts ( "loading symbols ..." );
    proceso -> LoadSymbols ( argv [ 3 ] );
  }
  else
  {
    printf ( "\n%s" , version );
    printf ( "\nCreated by Nicolas A. Economou ( neconomou@corest.com )" );
    printf ( "\nCore Security Technologies, Buenos Aires, Argentina ( 2008 )\n" );
    printf ( "\nsintax: iphonedbg <-e executable [arguments...]|-p pid executable>\n" );

    return ( 0 );
  }

/* Mientras el proceso no termine */
  while ( 1 )
  {
  /* Prompt */
    printf ( "%.8x> " , proceso -> GetPc () );

  /* Leo el comando desde el teclado */
    fgets ( buffer2 , 254 , stdin );

  /* Borro el new line */
    buffer2 [ strlen ( buffer2 ) - 1 ] = ';';
    buffer2 [ strlen ( buffer2 ) ] = '\x00';

  /* Cantidad de veces que tengo que ejecutar una linea por default */
    repeats = 1;

  /* Si tengo la directiva "repeat" adelante de los comandos */
    if ( strncmp ( buffer2 , "repeat" , 6 ) == 0 )
    {
    /* Tomo el primer comando */
      cmd = strtok ( buffer2 , ":" );

    /* Leo la cantidad de veces que tengo que repetir algo */
      sscanf ( cmd , "repeat %i" , &repeats );

    /* Achico el string */
      strcpy ( buffer2 , &cmd [ strlen ( cmd ) + 1 ] );
    }

  /* Repito una linea de comandos n veces */
    for ( cont = 0 ; cont < repeats ; cont ++ )
    {
    /* Hago una copia del buffer para poder parsearlo */
      strcpy ( buffer , buffer2 );

    /* Parseo el primer comando */
      cmd = strtok ( buffer , ";" );

    /* Si el comando es un ENTER */
      if ( cmd == NULL )
      {
      /* Obligo a que use el string vacio para que repita el comando anterior */
        buffer [ 0 ] = '\x00';
        cmd = buffer;
      }

    /* Proceso los comandos */
      while ( cmd != NULL )
      {
      /* Ejecuto el siguiente comando */
        if ( preprocess_command ( proceso , cmd ) == FALSE )
        {
        /* Si estoy attachado al proceso */
          if ( attachado == TRUE )
          {
          /* Para que el proceso siga corriendo */
            proceso -> Continue ();
          }
          else
          {
          /* Para que el proceso se cierre junto con el debugger */
            proceso -> Close ();
          }

        /* Cierro el debugger */
          return ( TRUE );
        }

      /* Separo el proximo comando ( no uso NULL porque strtok la uso mas abajo ) */
        cmd = strtok ( &cmd [ strlen ( cmd ) + 1 ] , ";" );
      }
    }
  }

  return ( FALSE );
}

/****************************************************************************/
/****************************************************************************/

/* Funciones */

/****************************************************************************/

int preprocess_command ( Debug *proceso , char *linea )
{
  char linea_duplicada [ 256 ];
  char buffer [ 256 ];
  char direccion_string [ 256 * 8 ];
  char *argv [ 256 ];
  _breakpoint *breakpoint;
  unsigned int len_register;
  unsigned int bytes_leidos;
  unsigned int addr_base;
  unsigned int addr_limite;
  unsigned int addr_found;
  unsigned int addr;
  static unsigned int direccion;
  static char comando [ 256 ];
  static int palabras_anteriores = 0;
  int *threads;
  int thread;
  int cantidad_threads;
  int palabras;
  int ret = TRUE;
  int pos;

/* Hago una copia de la linea de comandos */
  strcpy ( linea_duplicada , linea );

/* Parseo la linea de comandos */
  palabras = parsear_argumentos ( linea_duplicada , argv );

/* Por compatibilidad */
  if ( palabras >= 1 )
  {
    strcpy ( comando , argv [ 0 ] );
  }

/* Si tengo 2 palabras o mas */
  if ( palabras >= 2 )
  {
  /* Hago un backup del segundo parametro */
    strcpy ( direccion_string , argv [ 1 ] );

  /* Transformo el segundo parametro en una direccion */
    direccion = string_to_direccion ( proceso , direccion_string );
  }
/* Si es un ENTER, repito el comando anterior + direccion siguiente */
  else if ( palabras == 0 )
  {
  /* Repito la cantidad de palabras anteriores */
    palabras = palabras_anteriores;
  }

/* Si es imprimir la ayuda */
  if ( ( palabras == 1 ) && ( strcmp ( comando , "h" ) == 0 ) )
  {
    printf ( "\n" );
    printf ( "h                         -help\n" );
    printf ( "v                         -version\n" );
    printf ( "q                         -quit program\n" );
    printf ( "r [reg[=expression]]      -print or set registers\n" );
    printf ( "g [expression]            -run\n" );
//    printf ( "t [value]                 -trace execution n times\n" );
    printf ( "p                         -trace execution not entering to calls\n" );
    printf ( "u [expression]            -print code\n" );
    printf ( "db expression             -read byte format memory\n" );
    printf ( "dd expression             -read dword format memory\n" );
    printf ( "eb expression b1 b2 ...   -write byte format memory\n" );
    printf ( "bp expression [condition] -set breakpoint [reg{<|<=|==|>=|>|<>}value]\n" );
    printf ( "bc expression             -clear breakpoint\n" );
    printf ( "bl                        -show breakpoint list\n");
    printf ( "m                         -show memory map\n" );
    printf ( "s addr1 addr2 b1 b2 ...   -search from addr1 to addr2\n" );
    printf ( "f addr1 addr2 byte        -fill from addr1 to addr2 with byte value\n" );
    printf ( "x <*|lib>!<*|symbol>      -view symbols\n" );
    printf ( "~                         -threads list\n" );
    printf ( "~<0..n>r                  -show register values from the thread number\n" );
    printf ( "'enter'                   -repeat last command\n" );
    printf ( "\n" );
    printf ( "note: * to set breakpoints in thumb mode write address+1.\n" );
    printf ( "      * expression is converted to address.\n" );
    printf ( "      * to execute many cmds in a line use ';'.\n" );
    printf ( "      * to execute many times a line: ex. 'repeat 3:r;g;'.\n" );
    printf ( "\n" );
  }
/* Si es imprimir la version */
  else if ( ( palabras == 1 ) && ( strcmp ( comando , "v" ) == 0 ) )
  {
    printf ( "%s\n" , version );
  }
/* Si es continuar */
  else if ( strcmp ( comando , "g" ) == 0 )
  {
  /* Si tiene una direccion como parametro */
    if ( palabras == 2 )
    {
    /* Pongo un breakpoint one shot en esa direccion */
      if ( proceso -> SetOneShotBreakpoint ( direccion ) == FALSE )
      {
        puts ( "address error !!!" );
      }   
    }

  /* Ejecuto la instruccion go */
    ret = continue2 ( proceso );

  /* Tomo la direccion actual */
    direccion = proceso -> GetPc ();
  }
/* Si es single step */
//  else if ( strcmp ( comando , "t" ) == 0 )
//  {
//  /* Si no tiene parametros */
//    if ( palabras == 1 )
//    {
//      contador_steps = 1;
//    }
//    else
//    {
//      contador_steps = atoi ( argv [ 1 ] );
//    }
//
//  /* Habilito el trace flag */
//    proceso -> EnableTrace ();
//
//  /* Ejecuto la instruccion trace */
//    ret = continue2 ( proceso );
//
//  /* Tomo la direccion actual */
//    direccion = proceso -> GetPc ();
//
//  /* Deshabilito el trace flag */
//    proceso -> DisableTrace ();
//  }
/* Si es single step saltando los calls */
  else if ( strcmp ( comando , "p" ) == 0 )
  {
  /* traduzco la instruccion */
    bin2asm ( proceso , proceso -> GetPc () , &bytes_leidos );

  /* Pongo un breakpoint en la siguiente instruccion */
    proceso -> SetOneShotBreakpoint ( proceso -> GetPc () + bytes_leidos );

  /* Ejecuto la instruccion trace */
    ret = continue2 ( proceso );

  /* Tomo la direccion actual */
    direccion = proceso -> GetPc ();
  }
/* Si quiero ver el estado de los registros */
  else if ( strcmp ( comando , "r" ) == 0 )
  {
  /* Si quiero ver los registros */
    if ( palabras == 1 )
    {
    /* Muestro el codigo de donde estoy parado */
      imprimir_registros ( proceso , proceso -> GetCurrentTid () );
    }
  /* Si quiero ver el valor de un registro */
    else if ( ( palabras == 2 ) && ( is_register ( argv [ 1 ] , &len_register ) == TRUE ) )
    {
      printf ( "%s=%x\n" , argv [ 1 ] , get_register_value ( proceso , argv [ 1 ] ) );
    }
  /* Si quiero setear el valor de un registro */
    else if ( ( palabras == 4 ) && ( is_register ( argv [ 1 ] , &len_register ) == TRUE ) && ( strcmp ( argv [ 2 ] , "=" ) == 0 ) )
    {
    /* Seteo el valor del registro */
      if ( set_register_value ( proceso , argv [ 1 ] , argv [ 3 ] ) == FALSE )
      {
      /* Mensaje de ERROR */
        printf ( "cannot set register %s\n" , argv [ 1 ] );
      }
    }
    else
    {
      puts ( "register sintax error !!!" );
    }
  }
/* Si quiero ver el codigo desensamblado */
  else if ( strcmp ( comando , "u" ) == 0 )
  {
  /* Muestro el codigo de donde estoy parado */
    direccion = imprimir_codigo ( proceso , direccion );
  }
/* Si es finalizar el proceso */
  else if ( ( palabras == 1 ) && ( strcmp ( comando , "q" ) == 0 ) )
  {
    ret = FALSE;
  }
/* Si quiero ver una zona de memoria */
  else if ( ( palabras == 2 ) && ( strcmp ( comando , "db" ) == 0 ) )
  {
  /* Imprimo la memoria leida */
    direccion = imprimir_memoria_byte ( proceso , direccion , 128 );

  /* Dejo un espacio */
    printf ( "\n" );
  }
/* Si quiero ver una zona de memoria en formato dword */
  else if ( ( palabras == 2 ) && ( strcmp ( comando , "dd" ) == 0 ) )
  {
  /* Imprimo la memoria leida */
    direccion = imprimir_memoria_dword ( proceso , direccion , 128 / 4 );

  /* Dejo un espacio */
    printf ( "\n" );
  }
/* Si quiero escribir una zona de memoria */
  else if ( ( palabras >= 3 ) && ( strcmp ( comando , "eb" ) == 0 ) )
  {
  /* Direccion a donde escribir */
    sscanf ( argv [ 1 ] , "%x" , &addr_base );

  /* Transformo los caracteres ascii a binario */
    ascii_bytes_to_bytes ( buffer , &argv [ 2 ] , palabras - 2 );

  /* Si no pude escribir en el proceso */
    if ( proceso -> WriteMemory ( ( void * ) addr_base , buffer , palabras - 2 ) == FALSE )
    {
      printf ( "write error !!!\n" );
    }
  }
/* Si quiero poner un breakpoint */
  else if ( strcmp ( comando , "bp" ) == 0 )
  {
  /* Si es un breakpoint no condicional */
    if ( palabras == 2 )
    {
    /* Pongo un breakpoint en la direccion */
      if ( proceso -> SetBreakpoint ( direccion ) == FALSE )
      {
        puts ( "breakpoint error !!!" );
      }

    /* Creo un objeto breakpoint */
      breakpoint = ( _breakpoint * ) malloc ( sizeof ( _breakpoint ) );

    /* Seteo la direccion del breakpoint */
      breakpoint -> address = direccion;

    /* Seteo el breakpoint como no condicional */
      breakpoint -> conditional = FALSE;

    /* Agrego el breakpoint a la lista */
      breakpoints [ direccion & 0xfffffffe ] = breakpoint;
    }
    else if ( palabras == 3 )
    {
    /* Si condicion no esta bien */
      if ( isCorrectSintaxCondition ( argv [ 2 ] ) == TRUE )
      {
      /* Pongo un breakpoint en la direccion */
        if ( proceso -> SetBreakpoint ( direccion ) == FALSE )
        {
          puts ( "breakpoint error !!!" );
        }

      /* Creo un objeto breakpoint */
        breakpoint = ( _breakpoint * ) malloc ( sizeof ( _breakpoint ) );

      /* Seteo la direccion del breakpoint */
        breakpoint -> address = direccion;

      /* Seteo el breakpoint como condicional */
        breakpoint -> conditional = TRUE;

      /* Guardo la condicion */
        strncpy ( breakpoint -> condition , argv [ 2 ] , 31 );
        breakpoint -> condition [ strlen ( argv [ 2 ] ) ] = '\0';

      /* Agrego el breakpoint a la lista */
        breakpoints [ direccion & 0xfffffffe ] = breakpoint;
      }
      else
      {
        puts ( "breakpoint condition error !!!" );
      }
    }
    else
    {
      puts ( "breakpoint command error !!!" );
    }
  }
/* Si quiero eliminar un breakpoint */
  else if ( ( palabras == 2 ) && ( strcmp ( comando , "bc" ) == 0 ) )
  {
  /* Elimino el breakpoint */
    if ( proceso -> UnsetBreakpoint ( direccion ) == TRUE )
    {
    /* Leo la estructura del breakpoint */
      breakpoint = breakpoints [ direccion & 0xfffffffe ];

    /* Elimino el breakpoint de la lista */
      breakpoints.erase ( direccion & 0xfffffffe );

    /* Libero la memoria ocupada por el objeto */
      free ( breakpoint );
    }
    else
    {
      puts ( "breakpoint not found !!!" );
    }
  }
  else if ( ( palabras == 1 ) && ( strcmp ( comando , "bl" ) == 0 ) )
  {
  /* Entre llaves porque solo uso variables locales */
    {
      list <BreakpointInfo *> *user_breakpoints;
      list <BreakpointInfo *> :: iterator it_user_breakpoint;
      BreakpointInfo *user_breakpoint;

    /* Listo los breakpoints */
      user_breakpoints = proceso -> GetBreakpointList ();

    /* Cabezera */
      printf ( "address \tstatus \t\tshots\n" );

    /* Recorro todos los breakpoints */
      for ( it_user_breakpoint = user_breakpoints -> begin () ; it_user_breakpoint != user_breakpoints -> end () ; it_user_breakpoint ++ )
      {
      /* Levanto el siguiente breakpoint */
        user_breakpoint = *it_user_breakpoint;

      /* Imprimo el breakpoint */
        printf ( "%.8x\t%s\t\t%s\n" , user_breakpoint -> address , ( user_breakpoint -> enabled ) ? "ena":"dis" , ( user_breakpoint -> oneshot ) ? "one":"all" );
      }
    }
  }
/* Si quiero ver el mapa de memoria del proceso */
  else if ( ( palabras == 1 ) && ( strcmp ( comando , "m" ) == 0 ) )
  {
    {
      char permisos [ 256 ];
      unsigned int base;
      unsigned int len;
      unsigned int attributes;
      int pos = 0;
      int ret;

    /* Imprimo todo el mapa de memoria del proceso */
      while ( pos < 0xfffff000 )
      {
      /* Leo los atributos de la zona de memoria */
        ret = proceso -> GetMemoryAttributes ( pos , &base , &len , &attributes );

      /* Si la zona de memoria esta mapeada */
        if ( ret == TRUE )
        {
        /* Inicializo los permisos */
          strcpy ( permisos , "" );

        /* Transformo los permisos en strings */
          if ( attributes & DEBUG_PROTECTION_READ )
          {
            strcat ( permisos , "READ" );
          }

          if ( attributes & DEBUG_PROTECTION_WRITE )
          {
            strcat ( permisos , "|WRITE" );
          }

          if ( attributes & DEBUG_PROTECTION_EXECUTE )
          {
            strcat ( permisos , "|EXECUTE" );
          }

        /* Imprimo los permisos al usario */
          printf ( "%.8x %.8x %s\n" , base , len , permisos );

        /* Avanzo en la memoria lo que mide el bloque */
          pos = base + len;
        }
        else
        {
        /* Avanzo una pagina en la memoria */
          pos = pos + 0x1000;
        }
      }
    }
  }
/* Si quiero buscar un patron en la memoria */
  else if ( strcmp ( comando , "s" ) == 0 )
  {
  /* Si la cantidad de parametros es incorrecta */
    if ( palabras >= 4 )
    {
    /* Convierto los parametros en direcciones */
      sscanf ( argv [ 1 ] , "%x" , &addr_base );
      sscanf ( argv [ 2 ] , "%x" , &addr_limite );

    /* Si la direccion base no sobrepasa a la direccion limite */
      if ( addr_base < addr_limite )
      {
      /* Transformo todos los argumentos en bytes a buscar */
        ascii_bytes_to_bytes ( buffer , &argv [ 3 ] , palabras - 3 );

      /* Recorro toda la memoria buscando patrones */
        for ( addr = addr_base ; addr < addr_limite ; addr = addr_found + 1 )
        {
        /* Si encontre el patron en la memoria */
          if ( proceso -> Search ( ( void * ) addr , ( void * ) addr_limite , buffer , palabras - 3 , &addr_found ) == TRUE )
          {
          /* Imprimo la direccion encontrada */
            imprimir_memoria_byte ( proceso , addr_found , 16 );
          }
        /* Si no encontre nada */
          else
          {
          /* Corto la busqueda */
            break;
          }
        }
      }
      else
      {
        puts ( "address error !" );
      }
    }
    else
    {
      puts ( "sintax error !" );
    }
  }
/* Si quiero llenar una zona de memoria */
  else if ( strcmp ( comando , "f" ) == 0 )
  {
  /* Si la cantidad de parametros es incorrecta */
    if ( palabras == 4 )
    {
    /* Convierto los parametros en direcciones */
      sscanf ( argv [ 1 ] , "%x" , &addr_base );
      sscanf ( argv [ 2 ] , "%x" , &addr_limite );

    /* Si la direccion base no sobrepasa a la direccion limite */
      if ( addr_base < addr_limite )
      {
      /* Transformo todos los argumentos en bytes a buscar */
        ascii_bytes_to_bytes ( buffer , &argv [ 3 ] , 1 );

      /* Recorro toda la zona de memoria a escribir */
        for ( addr = addr_base ; addr < addr_limite ; addr ++ )
        {
        /* Escribo una byte en esa direccion */
          proceso -> WriteMemory ( ( void * ) addr , buffer , 1 );
        }
      }
      else
      {
        puts ( "address error !" );
      }
    }
    else
    {
      puts ( "sintax error !" );
    }
  }
/* Si quiero ver los simbolos */
  else if ( strcmp ( comando , "x" ) == 0 )
  {
  /* Si tengo 2 parametros */
    if ( palabras == 2 )
    {
    /* Imprimo todos los simbolos que matcheen */
      {
        char *libs_functions = argv [ 1 ];
        const char *symbol;
        unsigned int address;

      /* Averiguo el simbolo */
        while ( proceso -> GetSymbolFrom ( libs_functions , &address , &symbol ) == TRUE )
        {
        /* Imprimo el proximo simbolo que matcheo */
          printf ( "%.8x %s\n" , address , symbol );

        /* Para que no vuelva a pedir el primer simbolo */
          libs_functions = NULL;
        }
      }
    }
    else
    {
    /* Mensaje al usuario */
      printf ( "sintax: x lib!function\n" );
    }
  }
/* Si quiero ver la lista de threads */
  else if ( strncmp ( comando , "~" , 1 ) == 0 )
  {
  /* Si solo quiero ver la lista de threads */
    if ( strlen ( comando ) == 1 )
    {
    /* Pido la lista de threads */
      threads = proceso -> GetThreadList ( &cantidad_threads );
  
    /* Recorro toda la lista de threads */
      for ( pos = 0 ; pos < cantidad_threads ; pos ++ ) 
      {
      /* Thread siguiente */
        thread = threads [ pos ];

      /* Imprimo el thread ID */
        printf ( "%.2i%s %i\n" , pos , ( proceso -> GetCurrentTid () == thread ) ? ".":" " , thread );
      }
    }
  /* Si quiero ver los registros de algun thread en particular */
    else
    {
    /* Si solo escribi un numero y una letra 'r' */
      if ( ( sscanf ( comando + 1 , "%i%c%c" , &thread , &buffer [ 0 ] , &buffer [ 1 ] ) == 2 ) && ( buffer [ 0 ] == 'r' ) )
      {
      /* Pido la lista de threads */
        threads = proceso -> GetThreadList ( &cantidad_threads );

      /* Si el numero de thread no supera la cantidad de threads */
        if ( thread < cantidad_threads )
        {
        /* Imprimo los registros del thread elegido */
          imprimir_registros ( proceso , threads [ thread ] );
        }
      /* Si no existe el thread */
        else
        {
          puts ( "thread not exist !" );
        }
      }
      else
      {
        puts ( "sintax error !" );
      }
    }
  }
/* Si no es ningun comando conocido */
  else
  {
    puts ( "invalid command !!!" );
  }

/* Guardo la cantidad de parabras anteriores */
  palabras_anteriores = palabras;

  return ( ret );
}

/****************************************************************************/

int continue2 ( Debug *proceso )
{
  static int proceso_finalizado = FALSE;
  int salir = FALSE;
  int ret = TRUE;

/* Si el proceso finalizo */
  if ( proceso_finalizado == TRUE )
  {
    return ( FALSE );
  }

/* Comienzo a debugear */
  while ( salir == FALSE )
  {
  /* Continuo el proceso */
    proceso -> Continue ();

  /* Espero una seal del proceso */
    if ( proceso -> Wait () )
    {
    /* Vuelvo a la linea de comandos */
      salir = TRUE;

    /* Si es un ACCESS VIOLATION */
      if ( proceso -> GetEventCode () == DEBUG_EXCEPTION_ACCESS_VIOLATION )
      {
        printf ( "ACCESS VIOLATION\n" );
      }
    /* Si es un BREAKPOINT */
      else if ( proceso -> GetEventCode () == DEBUG_EXCEPTION_BREAKPOINT )
      {
      /* Si es un breakpoint condicional */
        if ( isConditionalBreakpoint ( proceso -> GetExceptionAddress () ) )
        {
        /* Si no se cumple la condicion */
          if ( isConditionalBreakpointTrue ( proceso ) == FALSE )
          {
          /* No vuelvo a la linea de comandos */
            salir = FALSE;
          }
          else
          {
            printf ( "BREAKPOINT\n" );
          }
        }
        else
        {
          printf ( "BREAKPOINT\n" );
        }
      }
    /* Si es un EXIT_PROCESS */
      else if ( proceso -> GetEventCode () == DEBUG_EXCEPTION_EXIT_PROCESS )
      {
        printf ( "EXIT PROCESS\n" );
        exit ( 0 );
      }
      else
      {
        printf ( "unknown exception: %i\n" , proceso -> GetEventCode () );
      }
    }
  }

/* Muestro los registros de donde estoy parado */
  imprimir_registros ( proceso , proceso -> GetCurrentTid () );

  return ( ret );
}

/****************************************************************************/

int parsear_argumentos ( char *linea , char *argv [] )
{
  int palabras = 0;

/* Nombre del comando */
  argv [ 0 ] = strtok ( linea , " " );

/* Si hay algun parametro */
  if ( argv [ 0 ] != NULL )
  {
  /* Recorro la linea de comandos */
    do
    {
    /* Incremento la cantidad de palabras encontradas */
      palabras ++;
    }
    while ( ( argv [ palabras ] = strtok ( NULL , " " ) ) != NULL );
  }

  return ( palabras );
}

/****************************************************************************/

unsigned int get_register_value ( Debug *proceso , char *registro )
{
  unsigned int registros [ CANTIDAD_REGISTROS ];
  unsigned int ret;
  unsigned int i;

/* Leo el estado de los registros */
  proceso -> GetRegisters ( proceso -> GetCurrentTid () , registros );

/* Retorno el valor de los registros */
  for ( i = 0 ; i < CANTIDAD_REGISTROS ; i ++ )
  {
  /* Si es el registro que estoy buscando */
    if ( strcmp ( register_names [ i ] , registro ) == 0 )
    {
    /* Valor del registro */
      ret = registros [ i ];

    /* Corto la busqueda */
      break;
    }
  }

  return ( ret );
}

/****************************************************************************/

int set_register_value ( Debug *proceso , char *registro , char *expresion )
{
  unsigned int registros [ CANTIDAD_REGISTROS ];
  char valor_a_transformar [ 256 * 4 ];
  unsigned int valor;
  int i;
  int ret = FALSE;

/* Hago una copia de la expresion a transformar */
  strcpy ( valor_a_transformar , expresion );

/* Transformo la expresion en un valor */
  valor = string_to_direccion ( proceso , valor_a_transformar );

/* Leo el estado de los registros */
  proceso -> GetRegisters ( proceso -> GetCurrentTid () , registros );

/* Retorno el valor de los registros */
  for ( i = 0 ; i < CANTIDAD_REGISTROS ; i ++ )
  {
  /* Si es el registro que estoy buscando */
    if ( strcmp ( register_names [ i ] , registro ) == 0 )
    {
    /* Valor del registro */
      registros [ i ] = valor;

    /* Restauro los registros */
      ret = proceso -> SetRegisters ( proceso -> GetCurrentTid () , registros );

    /* Corto la busqueda */
      break;
    }
  }

  return ( ret );
}

/****************************************************************************/

unsigned int string_to_direccion ( Debug *proceso , char *direccion )
{
  unsigned int ret = 0;

/* Reemplazo todos los nombres por direcciones */
  names_to_addresses ( proceso , direccion );

/* Reemplazo todos los registros por direcciones */
  registers_to_addresses ( proceso , direccion );

/* Realizo todas las sumas y restas de los numeros */
  addresses_to_address ( direccion );

/* direccion */
  sscanf ( direccion , "%x" , &ret );

  return ( ret );
}

/****************************************************************************/

int registers_to_addresses ( Debug *proceso , char *direccion )
{
  char registro [ 256 ];
  unsigned int len_register_name;
  int ret = TRUE;

/* Recorro todo el string buscando registros */
  while ( *direccion )
  {
  /* Si es un registro, lo traduzco a una direccion */
    if ( is_register ( direccion , &len_register_name ) )
    {
    /* Hago una copia del nombre del registro */
      strncpy ( registro , direccion , len_register_name );
      registro [ len_register_name ] = '\0';

    /* Traduzco el nombre del registro por el valor */
      sprintf ( registro , "%.8x" , register_to_address ( proceso , registro ) );

    /* Saco del string el nombre del registro */
      strsup ( direccion , len_register_name );

    /* Inserto dentro del string el valor del registro */
      strins ( direccion , registro );

    /* Incremento la posicion dentro del string */
      direccion = direccion + 7;
    }

  /* Incremento la posicion dentro del string */
    direccion ++;
  }

  return ( ret );
}

/****************************************************************************/

int names_to_addresses ( Debug *proceso , char *direccion )
{
  char name [ 256 + 1 ];
  char address_string [ 8 + 1 ];
  unsigned int address;
  int ret = TRUE;
  int len;

/* Recorro el string buscando nombres */
  while ( *direccion )
  {
  /* Longitud del substring */
    len = strcspn ( direccion , "+-" );

  /* Copio el substring a name */
    strncpy ( name , direccion , len );
    name [ len ] = '\0';

  /* Si pude resolver el nombre */
    if ( proceso -> ResolveNameToAddress ( name , &address ) == TRUE )
    {
    /* Elimino el nombre */
      strsup ( direccion , len );

    /* Trasnformo la direccion en un string */
      sprintf ( address_string , "%.8x" , address );

    /* Inserto la direccion del nombre dentro del string */
      strins ( direccion , address_string );

    /* Incremento mi posicion en la cadena */
      direccion = direccion + 8;
    }
    else
    {
    /* Incremento mi posicion en la cadena */
      direccion = direccion + len;
    }

  /* Si no llegue al final, sumo el signo a la posicion */
    if ( *direccion )
    {
      direccion ++;
    }
  }

  return ( ret );
}

/****************************************************************************/

int addresses_to_address ( char *direccion )
{
  char resultado [ 8 + 1 ];
  int n1, n2, n3;
  int ret = TRUE;

/* Aplico todas las operaciones matematicas */
  while ( 1 )
  {
  /* Si es una suma */
    if ( sscanf ( direccion , "%x+%x" , &n1 , &n2 ) == 2 )
    {
    /* Suma de los dos valores leidos */
      n3 = n1 + n2;
      sprintf ( resultado , "%x" , n3 );

    /* Saco del string los dos valores */
      strsup ( direccion , strcspn ( direccion , "+" ) + 1 );
      strsup ( direccion , strcspn ( direccion , "+-" ) );

    /* Inserto dentro del string el resultado */
      strins ( direccion , resultado );
    }
  /* Si es una resta */
    else if ( sscanf ( direccion , "%x-%x" , &n1 , &n2 ) == 2 )
    {
    /* Suma de los dos valores leidos */
      n3 = n1 - n2;
      sprintf ( resultado , "%x" , n3 );

    /* Saco del string los dos valores */
      strsup ( direccion , strcspn ( direccion , "-" ) + 1 );
      strsup ( direccion , strcspn ( direccion , "+-" ) );

    /* Inserto dentro del string el resultado */
      strins ( direccion , resultado );
    }
  /* Si es el fin */
    else if ( sscanf ( direccion , "%x+%x" , &n1 , &n2 ) == 1 )
    {
      break;
    }
  /* Si no pude reconocer la operacion matematica */
    else
    {
      ret = FALSE;
      break;
    }
  }

  return ( ret );
}

/****************************************************************************/

unsigned int register_to_address ( Debug *proceso , char *registro )
{
  unsigned int ret;

/* Averiguo el valor del registro */
  ret = get_register_value ( proceso , registro );

  return ( ret );
}

/****************************************************************************/

unsigned int selector_to_address ( Debug *proceso , char *registro )
{
  unsigned int selector_value;
  unsigned int ret;

/* Averiguo el valor del registro */
  selector_value = get_register_value ( proceso , registro );

/* Traduzco el selector en una direccion */
//  Debug_GetBaseSegment ( proceso , selector_value , &ret );

  return ( ret );
}

/****************************************************************************/

int is_register ( char *destino , unsigned int *len_register_name )
{
  int register_name_len;
  int cont;
  int ret = FALSE;

/* Averiguo la distancia hasta el proximo token */
  register_name_len = strpbrk ( destino , operadores ) - destino;

/* Si la longitud es cero retorno ERROR, ya que 'strncmp' con longitud cero da OK */
  if ( register_name_len == 0 )
  {
  /* Retorno ERROR */
    return ( FALSE );
  }

/* Recorro todos los registros */
  for ( cont = 0 ; cont < ( sizeof ( register_names ) / sizeof ( char * ) ) ; cont ++ )
  {
  /* Si es el registro que busco */
    if ( strncmp ( destino , register_names [ cont ] , register_name_len ) == 0 )
    {
    /* Seteo lo que ocupa el registro */
      *len_register_name = strlen ( register_names [ cont ] );

      ret = TRUE;
      break;
    }
  }

  return ( ret );
}

/****************************************************************************/

int isCorrectSintaxCondition ( char *condicion )
{
  char condition_struct [ 255 + 1 ];
  char registro [ 16 + 1 ];
  char comparacion [ 2 + 1 ];
  int operador_correcto = FALSE;
  unsigned int valor;
  unsigned int len;
  int len_registro;
  int ret = FALSE;
  char *operators;

/* Busco la posicion del operador */
  operators = strpbrk ( condicion , "<=>" );

/* Si tiene alguno de los caracteres de condicion */
  if ( operators != NULL )
  {
  /* Calculo la longitud del registro */
    len_registro = operators - condicion;

  /* Genero el string formateado para leer correctamente el registro a comparar */
    sprintf ( condition_struct , "%%%is%%%is%%x" , len_registro , ( strchr ( "<=>" , operators [ 1 ] ) == NULL ) ? 1 : 2 );
  }
  else
  {
    return ( ret );
  }

/* Si el operador ocupa 1 byte */
  if ( sscanf ( condicion , condition_struct , registro , comparacion , &valor ) == 3 )
  {
  /* Operador correcto */
    operador_correcto = TRUE;
  }
/* Si el operador ocupa 2 bytes */
  else if ( sscanf ( condicion , condition_struct , registro , comparacion , &valor ) )
  {
  /* Operador correcto */
    operador_correcto = TRUE;
  }

/* Si el operador es correcto */
  if ( operador_correcto == TRUE )
  {
  /* Si el primer parametro no es un registro */
    if ( is_register ( registro , &len ) == FALSE )
    {
      return ( ret );
    }

  /* Si son iguales */
    if ( strcmp ( comparacion , "<" ) == 0 )
    {
      ret = TRUE;
    }
    else if ( strcmp ( comparacion , "<=" ) == 0 )
    {
      ret = TRUE;
    }
    else if ( strcmp ( comparacion , "==" ) == 0 )
    {
      ret = TRUE;
    }
    else if ( strcmp ( comparacion , ">=" ) == 0 )
    {
      ret = TRUE;
    }
    else if ( strcmp ( comparacion , ">" ) == 0 )
    {
      ret = TRUE;
    }
    else if ( strcmp ( comparacion , "<>" ) == 0 )
    {
      ret = TRUE;
    }
  }

/* Retorno si es valido como condition */
  return ( ret );
}

/****************************************************************************/

int isConditionalBreakpoint ( unsigned int direccion )
{
  _breakpoint *breakpoint;
  int ret = FALSE;

/* Si el breakpoint esta en la lista */
  if ( breakpoints.find ( direccion & 0xfffffffe ) != breakpoints.end () )
  {
  /* Leo el breakpoint que acaba de saltar */
    breakpoint = breakpoints [ direccion & 0xfffffffe ];

  /* Si es un breakpoint condicional */
    if ( breakpoint -> conditional == TRUE )
    {
    /* Retorno OK */
      ret = TRUE;
    }
  }

  return ( ret );
}

/****************************************************************************/

int isConditionalBreakpointTrue ( Debug *proceso )
{
  char condition_struct [ 255 + 1 ];
  char registro [ 16 + 1 ];
  char comparacion [ 2 + 1 ];
  char *operators;
  _breakpoint *breakpoint;
  unsigned int len_registro;
  unsigned int direccion;
  unsigned int valor1;
  unsigned int valor2;
  int ret = FALSE;

/* Direccion donde ocurrio el breakpoint */
  direccion = proceso -> GetExceptionAddress ();

/* Si el breakpoint esta en la lista */
  if ( breakpoints.find ( direccion & 0xfffffffe ) != breakpoints.end () )
  {
  /* Leo el breakpoint que acaba de saltar */
    breakpoint = breakpoints [ direccion & 0xfffffffe ];

  /* Busco la posicion del operador */
    operators = strpbrk ( breakpoint -> condition , "<=>" );

  /* Si tiene alguno de los caracteres de condicion */
    if ( operators != NULL )
    {
    /* Calculo la longitud del registro */
      len_registro = operators - breakpoint -> condition;

    /* Genero el string formateado para leer correctamente el registro a comparar */
      sprintf ( condition_struct , "%%%is%%%is%%x" , len_registro , ( strchr ( "<=>" , operators [ 1 ] ) == NULL ) ? 1 : 2 );
    }

  /* Parseo la comparacion */
    sscanf ( breakpoint -> condition , condition_struct , registro , comparacion , &valor2 );

  /* Transformo el registro en un valor */  
    valor1 = register_to_address ( proceso , registro );

  /* Si son iguales */
    if ( ( strcmp ( comparacion , "<" ) == 0 ) && ( valor1 < valor2 ) )
    {
      ret = TRUE;
    }   
    else if ( ( strcmp ( comparacion , "<=" ) == 0 ) && ( valor1 <= valor2 ) )
    {
      ret = TRUE;
    }
    else if ( ( strcmp ( comparacion , "==" ) == 0 ) && ( valor1 == valor2 ) )
    {
      ret = TRUE;
    }
    else if ( ( strcmp ( comparacion , ">=" ) == 0 ) && ( valor1 >= valor2 ) )
    {    
      ret = TRUE;
    }
    else if ( ( strcmp ( comparacion , ">" ) == 0 ) && ( valor1 > valor2 ) )
    {
      ret = TRUE;
    }
    else if ( ( strcmp ( comparacion , "<>" ) == 0 ) && ( valor1 != valor2 ) )
    {
      ret = TRUE;
    }
  }

  return ( ret );
}

/****************************************************************************/

#define BYTES 16

unsigned int imprimir_memoria_byte ( Debug *proceso , unsigned int direccion , unsigned int bytes_a_imprimir )
{
  char valor_a_imprimir [ 2 + 1 ];
  unsigned int cont;
  unsigned int i;
  unsigned int bytes_leidos;
  unsigned char byte;

/* Imprimo cada linea de la memoria en formato byte */
  for ( cont = 0 ; cont < bytes_a_imprimir ; cont = cont + BYTES )
  {
  /* Imprimo la direccion de la memoria */
    printf ( "%.8x | " , direccion );

  /* Imprimo en hexa */
    for ( i = 0 ; i < BYTES ; i ++ )
    {
    /* Imprimo la cantidad que pedi */  
      if ( i < ( bytes_a_imprimir - cont ) )
      {
      /* Si este byte pudo ser leido */
        if ( proceso -> ReadMemory ( ( void * ) &byte , ( void * ) ( direccion + i ) , 1 , &bytes_leidos ) )
        {
          sprintf ( valor_a_imprimir , "%.2x" , byte );
        }
        else
        {
          strcpy ( valor_a_imprimir , "??" );
        }

      /* Imprimo el valor */
        printf ( "%s" , valor_a_imprimir );

      /* Si es la mitad de la columna */
        if ( i == 7 )
        {
          printf ( "-" );
        }
        else
        {
          printf ( " " );
        }
      }
      else
      {
        printf ( "   " );
      }
    }

  /* Espacio entre las 2 columnas */
    printf ( "| " );

  /* Imprimo en caracteres */  
    for ( i = 0 ; i < BYTES ; i ++ )
    {
    /* Por si no puedo leer el byte */
      byte = '?';

    /* Leo el byte a imprimir */
      proceso -> ReadMemory ( ( void * ) &byte , ( void * ) ( direccion + i ) , 1 , &bytes_leidos );

      if ( i < ( bytes_a_imprimir - cont ) )
      {
        printf ( "%c" , ( isgraph ( byte ) ) ? byte : '.' );
      }
      else
      {
        printf ( " " );
      }
    }

  /* Fin de linea */
    printf ( "\n" );

  /* Incremento la direccion a mostrar */
    direccion = direccion + BYTES;
  }

/* Retorno la ultima direccion leida + BYTES */
  return ( direccion );
}

/****************************************************************************/

unsigned int imprimir_memoria_dword ( Debug *proceso , unsigned int direccion , unsigned int dwords_a_imprimir )
{
  unsigned int bytes_leidos;
  unsigned int dword;
  unsigned int cont;
  unsigned int i;

/* Imprimo cada linea de la memoria en formato dword */
  for ( cont = 0 ; cont < dwords_a_imprimir ; cont = cont + ( BYTES / 4 ) )
  {
  /* Imprimo la direccion de la memoria */
    printf ( "%.8x | " , direccion );

  /* Imprimo en hexa los dwords */
    for ( i = 0 ; i < ( BYTES / 4 ) ; i ++ )
    {
    /* Imprimo la cantidad que pedi */  
      if ( i < ( dwords_a_imprimir - cont ) )
      {
      /* Si pude leer un dword del proceso */
        if ( proceso -> ReadMemory ( ( void * ) &dword , ( void * ) ( direccion + ( i * 4 ) ) , 4 , &bytes_leidos ) )
        {
          printf ( "%.8x " , dword );
        }
        else
        {
          printf ( "???????? " );
        }
      }
    }

  /* Fin de linea */
    printf ( "\n" );

  /* Incremento la direccion a mostrar */
    direccion = direccion + BYTES;
  }

/* Retorno la ultima direccion leida + BYTES */
  return ( direccion );
}

/****************************************************************************/

unsigned int imprimir_codigo ( Debug *proceso , unsigned int direccion )
{
  char buffer [ 256 ];
  char *instruccion;
  unsigned int bytes_leidos;
  unsigned int cont, cont2;
  unsigned int len;
  unsigned int pos;
  int offset;

/* Imprimo en que funcion estoy parado */
  if ( proceso -> ResolveAddressToName ( direccion , buffer , &offset ) )
  {
    printf ( "%s+%x:\n" , buffer , offset );
  }

/* Contador de bytes leidos */
  pos = 0;

/* Leo memoria del proceso */
  proceso -> ReadMemory ( ( void * ) buffer , ( void * ) ( direccion & 0xfffffffe ) , 256 , &len );

/* Imprimo 8 lineas de codigo */
  for ( cont = 0 ; cont < 8 ; cont ++ )
  {
  /* traduzco la instruccion */
    instruccion = bin2asm ( proceso , direccion , &bytes_leidos );

  /* Si me pase del buffer de la memoria */
    if ( ( pos + bytes_leidos ) > len )
    {
    /* Mensaje de error */
      puts ( "memory error !!!" );

    /* Salgo */
      return ( direccion + pos );
    }

  /* Si es un opcode invalido */
    if ( bytes_leidos == 0 )
    {
      bytes_leidos = 1;
    }

  /* Direccion donde esta la instruccion */
    printf ( "%x " , direccion );
             
  /* Imprimo los bytes utilizados por la instruccion */
    for ( cont2 = 0 ; cont2 < bytes_leidos ; cont2 ++ )
    {
      printf ( "%.2x" , ( unsigned char ) buffer [ pos + cont2 ] );
    }

  /* Imprimo la instruccion */
    printf ( "\t\t%s\n" , instruccion );

  /* Incremento la direccion base */
    direccion = direccion + bytes_leidos;

  /* Incremento el contador de bytes leidos */
    pos = pos + bytes_leidos;
  }

/* Retorno la ultima direccion leida + BYTES */
  return ( direccion );
}

/****************************************************************************/

void ascii_bytes_to_bytes ( char *buffer , char *argv [] , int bytes_count )
{
  int pos;

/* Transformo todos los ascii bytes en bytes */
  for ( pos = 0 ; pos < bytes_count ; pos ++ )
  {
  /* Transformo el siguiente byte */
    sscanf ( argv [ pos ] , "%x" , &buffer [ pos ] );
  }
}

/****************************************************************************/

void imprimir_registros ( Debug *proceso , int thread )
{
  unsigned int regs [ CANTIDAD_REGISTROS ];
  char buffer [ 256 + 1 ];
  char memoria [ 8 ];
  char *instruccion;
  unsigned int bytes_leidos;
  int offset;
  int cont;
  int i, j;
  int ret;

/* Leo los registros */
  proceso -> GetRegisters ( thread , regs );

/* Imprimo los registros */
  for ( i = 0 ; i < CANTIDAD_REGISTROS ; i = i + 4 )
  {
    for ( j = 0 ; ( ( i + j ) < CANTIDAD_REGISTROS ) && ( j < 4 ) ; j ++ )
    {
      printf ( "%s=%.8x\t" , register_names [ i + j ] , regs [ i + j ] );
    }

  /* Paso a la linea siguiente */
    printf ( "\n" );
  }

/* Si estoy imprimiendo los registros del current thread */
  if ( proceso -> GetCurrentTid () == thread )
  {
  /* Imprimo en que funcion estoy parado */
    if ( proceso -> ResolveAddressToName ( proceso -> GetPc () , buffer , &offset ) )
    { 
      printf ( "%s+%x:\n" , buffer , offset );
    }

  /* Imprimo el EIP */
    printf ( "pc=%.8x " , proceso -> GetPc () );

  /* Leo la memoria para mostrarla como instrucciones */
    ret = proceso -> ReadMemory ( memoria , ( void * ) proceso -> GetPc () , 4 , &bytes_leidos );

  /* traduzco la instruccion */
    instruccion = bin2asm ( proceso , proceso -> GetPc () , &bytes_leidos );

  /* Imprimo la memoria en forma hexadecimal */
    for ( cont = 0 ; cont < bytes_leidos ; cont ++ )
    {
      printf ( "%.2x " , ( unsigned char ) memoria [ cont ] );
    }

  /* Imprimo la instruccion */
    printf ( "%s" , ( ret == TRUE ) ? instruccion : "???" );

  /* Nueva linea */
    printf ( "\n\n" );
  }
}

/****************************************************************************/

char *bin2asm ( Debug *proceso , unsigned int direccion , unsigned int *bytes_interpretados )
{
  static char traduccion [ 256 ];

/* Si no pude desensamblar esa direccion */
  if ( proceso -> Disassembly ( direccion , bytes_interpretados , traduccion ) == FALSE )
  {
  /* Pongo incognitas como instruccion */
    strcpy ( traduccion , "???" );
  }

/* Retorno la instruccion desemsamblada */
  return ( traduccion );
}

/****************************************************************************/

char *strins ( char *dest , char *source )
{
  int len_source;
  int len_dest;

/* Longitud de la cadena a copiar */
  len_source = strlen ( source );

/* Longitud de la cadena a desplazar */
  len_dest = strlen ( dest );

/* Abro el lugar necesario en el destino para poder insertar la nueva cadena */
  memmove ( dest + len_source , dest , len_dest + 1 );

/* Inserto la nueva cadena */
  memcpy ( dest , source , len_source );

/* Retorno la posicion desplazada */
  return ( dest + len_source );
}

/****************************************************************************/

char *strsup ( char *cadena , int bytes_a_suprimir )
{
  int len_cadena;

/* Longitud de la subcadena a suprimir */
  len_cadena = strlen ( cadena );

/* Si la longitud de la subcadena es mayor que los bytes a suprimir */
  if ( len_cadena > bytes_a_suprimir )
  {
    strcpy ( cadena , cadena + bytes_a_suprimir );
  }
  else
  {
    cadena [ 0 ] = '\0';
  }

  return ( cadena );
}

/****************************************************************************/
/****************************************************************************/

